import argparse
from util.model_util import *
from util.cost_util import *
from util.data_processing import *


def train(args):
    # load data
    if args.approach == 'PDL':
        train_dedicated_samples = 'prim'
    elif args.approach == 'DLLCN':
        train_dedicated_samples = '3GP' if not args.use_4GP else '4GP'
    else:
        assert(False)
    data = Dataset(args.dataset, scale_tonal=True, train_dedicated_samples=train_dedicated_samples, n_test=args.test_size, seed_for_test_set=789)
    data.load_data()

    # model config
    model_train_config = ModelTrainConfig(args.dataset, data, args.approach)
    model_train_config.ep = args.ep

    # training...
    for train_size in args.train_sizes:
        print("\n")
        print("=" * 20 + f" train_size = {train_size}" + "=" * 20)
        if train_size < len(data.train_dedicated_idx):
            print(f"train_size ({train_size}) is smaller than # of training-dedicated samples, i.e. {len(data.train_dedicated_idx)}. Skip this train_size.")
            continue
        n_valid = int(train_size * 0.05)
        n_train = train_size - n_valid

        my_train(
            X=data.get_input(),
            y=data.get_output(),
            xval_polyFeat=False,
            ns=model_train_config.ns,
            activation='relu',
            BN=True,
            dropout=model_train_config.dropout,
            loss=model_train_config.loss_fn,
            init_lr=model_train_config.lr, ep=model_train_config.ep,
            metricOn='valid',
            callbacks=model_train_config.callbacks,
            show_history=False,
            show_metric=True,
            show_metric_avg=True,
            xval_ratio=n_train,
            xval_ratioForValSet=n_valid,
            xval_ratioForTestSet=None,
            xval_seed=789,
            xval_folds=args.folds,
            xval_saveAllModels=False,
            model_outputs=model_train_config.output_descr,
            idx_dedicated_for_training=data.train_dedicated_idx,
            kernel_initializer=model_train_config.weight_initializer,
            lrelu_alpha=model_train_config.lrelu_alpha,
            activation_final=model_train_config.activation_final,
            sigmoid_hshift_amount=None,
            bs=model_train_config.bs,
            CMYKG2cmykg=model_train_config.CMYKG2cmykg,
            cmykg_sigmoidAfterMul=model_train_config.cmykg_sigmoidAfterMul,
            cmykg_clipAfterMul=model_train_config.cmykg_clipAfterMul,
            interp_points=data.get_CN_ref_points(),
            interp_values=data.get_CN_ref_values(),
            hlayer_split=model_train_config.hlayer_split,
            ns2=model_train_config.ns2,
            n_out2=model_train_config.n_out2,
            n_out=model_train_config.n_out,
            deterministicSplit=True,
            test_set=data.split_test_idx(),
            idx_hiPrio_for_P=data.get_hiPrio_idx_for_P(),
            gamma_stratified_sampling=data.gamma_stratified_sampling,
            distance_based_P=False,
            set_sameSeed_P=True,
            set_seed_trainValidSplit=True,
            ensemble_thr_folds=True,
            run_eagerly = False, verbose=0)


def main(args):
    train_sizes_dict = {
        'stratasys': {
            'DLLCN': [300, 400, 500, 576, 700, 1024, 1280, 1600, 2000, 2500],
            'PDL': [243, 300, 350, 400, 576, 1024, 1280, 1600, 2000, 2500]
        },
        'mimaki_1': {
            'PDL': [200, 700, 1200, 1700, 2200, 2700]
        },
        'mimaki_2': {
            'DLLCN': [300, 500, 800, 1296, 1700, 2058, 2401],
            'PDL': [256, 500, 800, 1296, 1700, 2058, 2401]
        },
    }
    assert (args.dataset in train_sizes_dict), f'Only datasets in {list(train_sizes_dict.keys())} are supported! '
    assert (args.approach in train_sizes_dict[args.dataset]), f'For {args.dataset}, only approaches in {list(train_sizes_dict[args.dataset].keys())} are supported! '
    if args.use_4GP:
        assert(args.dataset == 'mimaki_2'), '"--use_4GP" is only supported for mimaki_2!'
        assert(args.approach == 'DLLCN'), '"--use_4GP" can only be used with DLLCN!'
    if args.train_sizes is None:
        args.train_sizes = train_sizes_dict[args.dataset][args.approach]
        print(f"train_sizes is automatically set to {args.train_sizes}")

    print("\n" + "=" * 20 + " The training starts " + "=" * 20)
    print(datetime.now())

    train(args)



if __name__ == "__main__":
    parser = argparse.ArgumentParser(add_help=False, description='This is the code to reproduce the results of our '
                                                                 'submitted paper')
    optional = parser.add_argument_group('optional arguments')
    required = parser.add_argument_group('required arguments')
    optional.add_argument(
        '-h',
        '--help',
        action='help',
        default=argparse.SUPPRESS,
        help='show this help message and exit'
    )
    required.add_argument('--dataset', required=True, help='stratasys/mimaki_1/mamaki_2')
    required.add_argument('--approach', required=True, help='DLLCN/PDL')
    required.add_argument('--use_4GP', default=False, action='store_true', help='True/False. Whether to use 4GP for DLLCN, instead of 3GP')
    optional.add_argument('--train_sizes', type=int, nargs="*", default=None, help='Example: "--train_sizes 300 400 500". If not set, it will be automatically determined as per the submitted paper.')
    required.add_argument('--test_size', type=int, default=300, help='#samples in test set. Example: --test_set 300')
    required.add_argument('--folds', type=int, default=10, help='#times to split P into train/valid')
    required.add_argument('--ep', type=int, default=15000, help='#epochs of training')

    args = parser.parse_args()

    timestr = time.strftime("%Y%m%d-%H%M%S")
    log_fname = f"./results/{args.dataset}_{args.approach}" + ("_4GP" if args.use_4GP else "") + f"_{timestr}.txt"
    if not os.path.exists(os.path.dirname(log_fname)):
        os.mkdir(os.path.dirname(log_fname))
    # redirect stdout to the log file
    log_f = open(log_fname, 'w')
    print(f"The results will be continuously logged into file {log_fname}")
    sys.stdout.flush()
    time.sleep(5)
    sys.stdout = log_f

    print("=" * 20 + " Arguments " + "=" * 20)
    for arg in vars(args):
        print(arg, ":", getattr(args, arg))

    main(args)

    sys.stdout = sys.__stdout__
    log_f.close()

    print("\nALL DONE.", datetime.now())
    print(f"All the results can be found in file {log_fname}.")
