import tensorflow as tf
'''
import tensorflow
if tensorflow.__version__.startswith("1."):
    import tensorflow as tf
else:
    import tensorflow.compat.v1 as tf
    tf.disable_v2_behavior()
'''

from tensorflow.keras import backend as K
import tensorflow_probability as tfp

import numpy as np
from util.data_processing import sigmoid
from util.color_util import color_diff, d50, LightingMatrix_2Degree, get_lab_from_spectral
import util.color_constants as color_constants
import matplotlib.pyplot as plt
import sys
import copy
from scipy import interpolate

def my_loss_fn_decorator(func, loss):
    def wrapper(*args, **kwargs):
        kwargs['loss'] = loss
        return func(*args, **kwargs)

    return wrapper

def MSE(pred, gt):
    assert(pred.shape == gt.shape)
    return np.mean(np.square(pred-gt), axis=-1)

def RMSE(pred, gt):
    assert(pred.shape == gt.shape), print(pred.shape, gt.shape)
    return np.sqrt(np.mean(np.square(pred-gt), axis=-1))

def specRMSE(pred, gt):
    assert(pred.shape[-1] == 31)
    return RMSE(pred, gt)

def norm_mse(y_true, y_pred):
    print('==>norm_mse')
    norm = K.clip(K.abs(y_true), K.epsilon(), None)
    # norm = K.abs(K.concatenate((y_true[:,0:1], y_true[:,0:1], y_true[:,0:1]), axis=1)) + 1e-5
    # norm = K.clip(K.square(K.concatenate((y_true[:,0:1], y_true[:,0:1], y_true[:,0:1]), axis=1)), K.epsilon(), None)
    res = K.mean(K.square(y_pred - y_true) / norm, axis=-1)
    return res


def radian2degree(rad):
    pi_on_180 = 0.017453292519943295
    return rad / pi_on_180

def degree2radian(deg):
    pi_on_180 = 0.017453292519943295
    return deg * pi_on_180

# FIXME: if the input lab is out of range which is likely due to that upstream spectral layer has out-of-range
#  (normal range: [0.0, 1.0]) values that occurs typically for large learning rate, then this my_delta_e_cie2000() func
#  could yield inf or nan result. Possible solutions include: 1) bound the output range of spec layer by sigmoid
#  activation func or by clipping; 2) replace this loss func with others like euclidean distance loss
def my_delta_e_cie2000(lab_color_vector, lab_color_matrix, Kl=1, Kc=1, Kh=1, loss_mode=1):
    """
    Calculates the Delta E (CIE2000) of two colors.
    """
    L, a, b = lab_color_vector[:,0], lab_color_vector[:,1], lab_color_vector[:,2]

    avg_Lp = (L + lab_color_matrix[:, 0]) / 2.0

    C1 = K.sqrt(K.sum(K.pow(lab_color_vector[:, 1:], 2), axis=1))
    C2 = K.sqrt(K.sum(K.pow(lab_color_matrix[:, 1:], 2), axis=1))

    avg_C1_C2 = (C1 + C2) / 2.0

    G = 0.5 * (
        1
        - K.sqrt(
            K.pow(avg_C1_C2, 7.0)
            / (K.pow(avg_C1_C2, 7.0) + K.pow(25.0, 7.0))
        )
    )

    a1p = (1.0 + G) * a
    a2p = (1.0 + G) * lab_color_matrix[:, 1]

    C1p = K.sqrt(K.pow(a1p, 2) + K.pow(b, 2))
    C2p = K.sqrt(K.pow(a2p, 2) + K.pow(lab_color_matrix[:, 2], 2))

    avg_C1p_C2p = (C1p + C2p) / 2.0

    h1p = radian2degree(tf.math.atan2(b, a1p)) # K.degrees(K.atan2(b, a1p))
    #h1p += K.cast(h1p < 0, "float") * 360  # (h1p < 0) * 360
    h1p += K.cast(h1p < 0, h1p.dtype) * 360  # (h1p < 0) * 360

    h2p = radian2degree(tf.math.atan2(lab_color_matrix[:, 2], a2p)) # K.degrees(K.atan2(lab_color_matrix[:, 2], a2p))
    #h2p += K.cast(h2p < 0, "float") * 360
    h2p += K.cast(h2p < 0, h2p.dtype) * 360

    #avg_Hp = ((K.cast(K.abs(h1p - h2p) > 180, "float") * 360) + h1p + h2p) / 2.0
    avg_Hp = ((K.cast(K.abs(h1p - h2p) > 180, h1p.dtype) * 360) + h1p + h2p) / 2.0

    T = (
        1
        - 0.17 * K.cos(degree2radian(avg_Hp - 30))
        + 0.24 * K.cos(degree2radian(2 * avg_Hp))
        + 0.32 * K.cos(degree2radian(3 * avg_Hp + 6))
        - 0.2 * K.cos(degree2radian(4 * avg_Hp - 63))
    )

    diff_h2p_h1p = h2p - h1p
    #delta_hp = diff_h2p_h1p + K.cast(K.abs(diff_h2p_h1p) > 180, "float") * 360
    delta_hp = diff_h2p_h1p + K.cast(K.abs(diff_h2p_h1p) > 180, diff_h2p_h1p.dtype) * 360
    #delta_hp -= K.cast(h2p > h1p, "float") * 720
    delta_hp -= K.cast(h2p > h1p, h1p.dtype) * 720

    delta_Lp = lab_color_matrix[:, 0] - L
    delta_Cp = C2p - C1p
    delta_Hp = 2 * K.sqrt(C2p * C1p) * K.sin(degree2radian(delta_hp) / 2.0)

    S_L = 1 + (
        (0.015 * K.pow(avg_Lp - 50, 2))
        / K.sqrt(20 + K.pow(avg_Lp - 50, 2.0))
    )
    S_C = 1 + 0.045 * avg_C1p_C2p
    S_H = 1 + 0.015 * avg_C1p_C2p * T

    delta_ro = 30 * K.exp(-(K.pow(((avg_Hp - 275) / 25), 2.0)))
    R_C = K.sqrt(
        (K.pow(avg_C1p_C2p, 7.0))
        / (K.pow(avg_C1p_C2p, 7.0) + K.pow(25.0, 7.0))
    )
    R_T = -2 * R_C * K.sin(2 * degree2radian(delta_ro))

    Elabs = K.sqrt(
                K.pow(delta_Lp / (S_L * Kl), 2)
                + K.pow(delta_Cp / (S_C * Kc), 2)
                + K.pow(delta_Hp / (S_H * Kh), 2)
                + R_T * (delta_Cp / (S_C * Kc)) * (delta_Hp / (S_H * Kh))
            )
    if loss_mode == 1:
        return K.mean(Elabs)
    elif loss_mode == 2:
        return K.mean(K.square(Elabs))
    elif loss_mode == 3:
        return Elabs


def my_delta_e_cie2000_noMean(lab_color_vector, lab_color_matrix, Kl=1, Kc=1, Kh=1):
    return my_delta_e_cie2000(lab_color_vector, lab_color_matrix, Kl=Kl, Kc=Kc, Kh=Kh, loss_mode=3)


def LABfromSpectra_tf(spectra, illuminant='d50', observer='2'):
    assert (illuminant == 'd50')
    illuminant_arr = d50

    # Get XYZ -- from Tejas' MATLAB implementation
    Lh = np.transpose(LightingMatrix_2Degree(illuminant_arr)).astype('float32')
    XYZ = tf.linalg.matmul(spectra, Lh)

    # Get LAB -- modified based on colormath's implementation
    illum = color_constants.ILLUMINANTS[observer][illuminant]
    XYZ_tmp = tf.math.divide(XYZ, illum)
    #XYZ_tmp = tf.where(XYZ_tmp > color_constants.CIE_E, tf.math.pow(XYZ_tmp, (1.0 / 3.0)), (7.787 * XYZ_tmp) + (16.0 / 116.0))
    XYZ_tmp = tf.where(XYZ_tmp > color_constants.CIE_E, (tf.cast(XYZ_tmp > 0, XYZ_tmp.dtype) - 0.5) * 2 * tf.math.pow(tf.math.abs(XYZ_tmp), (1.0 / 3.0)), (7.787 * XYZ_tmp) + (16.0 / 116.0))
    temp_x, temp_y, temp_z = XYZ_tmp[:, 0], XYZ_tmp[:, 1], XYZ_tmp[:, 2]

    lab_l = (116.0 * temp_y) - 16.0
    lab_a = 500.0 * (temp_x - temp_y)
    lab_b = 200.0 * (temp_y - temp_z)

    # LABs = tf.stack([tf.reshape(lab_l, [-1, 1]), tf.reshape(lab_a, [-1, 1]), tf.reshape(lab_b, [-1, 1])], axis=1)
    LABs = tf.stack([lab_l, lab_a, lab_b], axis=1)
    return LABs

def deltaAlpha(alpha1, alpha2):
    return tf.math.abs(alpha1 - alpha2)

def scaleUpLab_tf(lab):
    '''
    lab = copy.deepcopy(lab0)
    lab[:, 0] *= 100
    lab[:, 1] = (lab[:, 1] - 0.5) * (128 * 2)
    '''
    a1 = tf.ones_like(lab[:,0:1]) * 0.
    a2 = tf.ones_like(lab[:,0:1]) * 0.5
    a3 = tf.ones_like(lab[:,0:1]) * 0.5
    a = tf.concat([a1,a2,a3], 1)
    print(a.shape)
    b1 = tf.ones_like(lab[:,0:1]) * 100.
    b2 = tf.ones_like(lab[:,0:1]) * (128*2)
    b3 = tf.ones_like(lab[:,0:1]) * (128*2)
    b = tf.concat([b1,b2,b3], 1)
    return (lab - a) * b

def ciede2000ScaledLAB(lab_and_alpha_arr1, lab_and_alpha_arr2, Kl=1, Kc=1, Kh=1):
    lab_arr1, alpha1 = scaleUpLab_tf(lab_and_alpha_arr1[:,:3]), lab_and_alpha_arr1[:,3]
    lab_arr2, alpha2 = scaleUpLab_tf(lab_and_alpha_arr2[:,:3]), lab_and_alpha_arr2[:,3]
    Elab = my_delta_e_cie2000_noMean(lab_arr1, lab_arr2, Kl=Kl, Kc=Kc, Kh=Kh)
    return Elab


# the pool of loss functions that are primitive (e.g., MSE, RMSE, euclidean distance, abs error)
class PrimitiveLossPool:
    @staticmethod
    def MSE_tensor(gt, pred):
        return K.mean(K.square(gt - pred), axis=-1)
    # add tiny epsilon before calling sqrt, otherwise CMYKG2cmykg model get NaN gradients, ruining entire model
    @classmethod
    def RMSE_tensor(cls, gt, pred, epsilon=1e-16):
        return K.sqrt(K.mean(K.square(gt - pred), axis=-1) + epsilon)
    @classmethod
    def RMSESquare_tensor(cls, gt, pred):  # = MSE
        return K.square(cls.RMSE_tensor(gt, pred))
    @classmethod
    def RMSEby250Square_tensor(cls, gt, pred):  # observed spec RMSE mean ~=0.004 which multiplies by 250 yielding 1.0
        return K.square(cls.RMSE_tensor(gt, pred) * 250)
    @classmethod
    def RMSECubic_tensor(cls, gt, pred):
        return K.pow(cls.RMSE_tensor(gt, pred), 3)
    @classmethod
    def RMSEby250Cubic_tensor(cls, gt, pred):
        return K.pow(cls.RMSE_tensor(gt, pred) * 250, 3)
    @classmethod
    def RMSEQuartic_tensor(cls, gt, pred):
        return K.pow(cls.RMSE_tensor(gt, pred), 4)
    @classmethod
    def RMSEby250Quartic_tensor(cls, gt, pred):
        return K.pow(cls.RMSE_tensor(gt, pred) * 250, 4)
    @staticmethod
    def MAE_tensor(gt, pred):
        return tf.math.reduce_mean(tf.math.abs(gt - pred), axis=-1)
    @staticmethod
    def by250CSE_tensor(gt, pred): # observed spec RMSE
        return K.mean(K.pow((gt - pred)*250, 3), axis=-1)
    @staticmethod
    def euclidean_distance_tensor(gt, pred, epsilon=1e-16):
        return K.sqrt(K.sum(K.square(gt - pred), axis=-1) + epsilon)
    @staticmethod
    def boundary_linear_tensor(pred):  # penalty on exceeding each side of the expected boundary of [0.0, 1.0]
        return tf.math.reduce_mean(tf.math.maximum(0., pred - 1.0) + tf.math.maximum(0., -pred))
    @staticmethod
    def boundary_square_tensor(pred):  # penalty on exceeding each side of the expected boundary of [0.0, 1.0]
        return tf.math.reduce_mean(tf.math.square(tf.math.maximum(0., pred - 1.0) + tf.math.maximum(0., -pred)))
    @staticmethod
    def allOnes_tensor(pred):  # penalty on all 1s prediction
        all_1_th = 1.0 - 1e-3
        all_01_bits_th = 16
        n_1_pred = tf.math.reduce_sum(tf.cast(tf.math.greater_equal(pred, all_1_th), pred.dtype), axis=-1)
        return tf.cast(tf.math.greater_equal(n_1_pred, all_01_bits_th), pred.dtype)  # 0.0 if no many ones otherwise 1.0
    @staticmethod
    def meanOne_tensor(pred):  # penalty on all 1s prediction
        mean_1_th = 1.0 - 1e-3
        mean_pred = tf.math.reduce_mean(pred, axis=-1)
        return tf.cast(tf.math.greater_equal(mean_pred, mean_1_th), pred.dtype)
    @staticmethod
    def mean_tensor(pred):  # penalty on all 1s prediction
        mean_pred = tf.math.reduce_mean(pred, axis=-1)
        return mean_pred
    @staticmethod
    def MAE_tensor(gt, pred):
        #return tf.math.reduce_mean(tf.math.abs(gt - pred), axis=-1)
        return tf.math.abs(gt - pred)
    @staticmethod
    def by250CSE_tensor(gt, pred): # observed spec RMSE
        return K.mean(K.pow((gt - pred)*250, 3), axis=-1)
    @staticmethod
    def boundary_linear_tensor(pred):  # penalty on exceeding each side of the expected boundary of [0.0, 1.0]
        return tf.math.reduce_mean(tf.math.maximum(0., pred - 1.0) + tf.math.maximum(0., -pred))
    @staticmethod
    def boundary_square_tensor(pred):  # penalty on exceeding each side of the expected boundary of [0.0, 1.0]
        return tf.math.reduce_mean(tf.math.square(tf.math.maximum(0., pred - 1.0) + tf.math.maximum(0., -pred)))
    @staticmethod
    def allOnes_tensor(pred):  # penalty on all 1s prediction
        all_1_th = 1.0 - 1e-3
        all_01_bits_th = 16
        n_1_pred = tf.math.reduce_sum(tf.cast(tf.math.greater_equal(pred, all_1_th), pred.dtype), axis=-1)
        return tf.cast(tf.math.greater_equal(n_1_pred, all_01_bits_th), pred.dtype)  # 0.0 if no many ones otherwise 1.0
    @staticmethod
    def meanOne_tensor(pred):  # penalty on all 1s prediction
        mean_1_th = 1.0 - 1e-3
        mean_pred = tf.math.reduce_mean(pred, axis=-1)
        return tf.cast(tf.math.greater_equal(mean_pred, mean_1_th), pred.dtype)
    @staticmethod
    def mean_tensor(pred):  # penalty on all 1s prediction
        mean_pred = tf.math.reduce_mean(pred, axis=-1)
        return mean_pred


class ColorDiff:
    @staticmethod
    def radian2degree(rad):
        pi_on_180 = 0.017453292519943295
        return rad / pi_on_180

    @staticmethod
    def degree2radian(deg):
        pi_on_180 = 0.017453292519943295
        return deg * pi_on_180

    # FIXME: if the input lab is out of range which is likely due to that upstream spectral layer has out-of-range
    #  (normal range: [0.0, 1.0]) values that occurs typically for large learning rate, then this my_delta_e_cie2000() func
    #  could yield inf or nan result. Possible solutions include: 1) bound the output range of spec layer by sigmoid
    #  activation func or by clipping; 2) replace this loss func with others like euclidean distance loss
    @classmethod
    def ciede2000_tensor(cls, lab_color_vector, lab_color_matrix, Kl=1, Kc=1, Kh=1, loss_mode=1, epsilon=0.0):
        """
        Calculates the Delta E (CIE2000) of two colors.
        """
        L, a, b = lab_color_vector[:,0], lab_color_vector[:,1], lab_color_vector[:,2]

        avg_Lp = (L + lab_color_matrix[:, 0]) / 2.0

        C1 = K.sqrt(K.sum(K.pow(lab_color_vector[:, 1:], 2), axis=1) + epsilon)
        C2 = K.sqrt(K.sum(K.pow(lab_color_matrix[:, 1:], 2), axis=1) + epsilon)

        avg_C1_C2 = (C1 + C2) / 2.0

        G = 0.5 * (
                1
                - K.sqrt(
            K.pow(avg_C1_C2, 7.0)
            / (K.pow(avg_C1_C2, 7.0) + K.pow(25.0, 7.0))
            + epsilon
        )
        )

        a1p = (1.0 + G) * a
        a2p = (1.0 + G) * lab_color_matrix[:, 1]

        C1p = K.sqrt(K.pow(a1p, 2) + K.pow(b, 2) + epsilon)
        C2p = K.sqrt(K.pow(a2p, 2) + K.pow(lab_color_matrix[:, 2], 2) + epsilon)

        avg_C1p_C2p = (C1p + C2p) / 2.0

        h1p = cls.radian2degree(tf.math.atan2(b, a1p)) # K.degrees(K.atan2(b, a1p))
        #h1p += K.cast(h1p < 0, "float") * 360  # (h1p < 0) * 360
        h1p += K.cast(h1p < 0, h1p.dtype) * 360  # (h1p < 0) * 360

        h2p = cls.radian2degree(tf.math.atan2(lab_color_matrix[:, 2], a2p)) # K.degrees(K.atan2(lab_color_matrix[:, 2], a2p))
        #h2p += K.cast(h2p < 0, "float") * 360
        h2p += K.cast(h2p < 0, h2p.dtype) * 360

        #avg_Hp = ((K.cast(K.abs(h1p - h2p) > 180, "float") * 360) + h1p + h2p) / 2.0
        avg_Hp = ((K.cast(K.abs(h1p - h2p) > 180, h1p.dtype) * 360) + h1p + h2p) / 2.0

        T = (
                1
                - 0.17 * K.cos(cls.degree2radian(avg_Hp - 30))
                + 0.24 * K.cos(cls.degree2radian(2 * avg_Hp))
                + 0.32 * K.cos(cls.degree2radian(3 * avg_Hp + 6))
                - 0.2 * K.cos(cls.degree2radian(4 * avg_Hp - 63))
        )

        diff_h2p_h1p = h2p - h1p
        #delta_hp = diff_h2p_h1p + K.cast(K.abs(diff_h2p_h1p) > 180, "float") * 360
        delta_hp = diff_h2p_h1p + K.cast(K.abs(diff_h2p_h1p) > 180, diff_h2p_h1p.dtype) * 360
        #delta_hp -= K.cast(h2p > h1p, "float") * 720
        delta_hp -= K.cast(h2p > h1p, h2p.dtype) * 720

        delta_Lp = lab_color_matrix[:, 0] - L
        delta_Cp = C2p - C1p
        delta_Hp = 2 * K.sqrt(C2p * C1p + epsilon) * K.sin(cls.degree2radian(delta_hp) / 2.0)

        S_L = 1 + (
                (0.015 * K.pow(avg_Lp - 50, 2))
                / K.sqrt(20 + K.pow(avg_Lp - 50, 2.0) + epsilon)
        )
        S_C = 1 + 0.045 * avg_C1p_C2p
        S_H = 1 + 0.015 * avg_C1p_C2p * T

        delta_ro = 30 * K.exp(-(K.pow(((avg_Hp - 275) / 25), 2.0)))
        R_C = K.sqrt(
            (K.pow(avg_C1p_C2p, 7.0))
            / (K.pow(avg_C1p_C2p, 7.0) + K.pow(25.0, 7.0))
            + epsilon
        )
        R_T = -2 * R_C * K.sin(2 * cls.degree2radian(delta_ro))

        Elabs = K.sqrt(
            K.pow(delta_Lp / (S_L * Kl), 2)
            + K.pow(delta_Cp / (S_C * Kc), 2)
            + K.pow(delta_Hp / (S_H * Kh), 2)
            + R_T * (delta_Cp / (S_C * Kc)) * (delta_Hp / (S_H * Kh))
            + epsilon
        )
        if loss_mode == 1:
            return K.mean(Elabs)
        elif loss_mode == 2:
            return K.mean(K.square(Elabs))
        elif loss_mode == 3:
            return Elabs
    @classmethod
    def ciede2000MSE_tensor(cls, lab_color_vector, lab_color_matrix, Kl=1, Kc=1, Kh=1):
        return cls.ciede2000_tensor(lab_color_vector, lab_color_matrix, Kl=Kl, Kc=Kc, Kh=Kh, loss_mode=2, epsilon=1e-16)
    @classmethod
    def ciede2000noMean_tensor(cls, lab_color_vector, lab_color_matrix, Kl=1, Kc=1, Kh=1):
        return cls.ciede2000_tensor(lab_color_vector, lab_color_matrix, Kl=Kl, Kc=Kc, Kh=Kh, loss_mode=3, epsilon=1e-16)
    @classmethod
    def ciede2000noMeanSquare_tensor(cls, lab_color_vector, lab_color_matrix, Kl=1, Kc=1, Kh=1):
        return tf.math.square(cls.ciede2000_tensor(lab_color_vector, lab_color_matrix, Kl=Kl, Kc=Kc, Kh=Kh, loss_mode=3, epsilon=1e-16))
    @classmethod
    def ciede2000noMeanCubic_tensor(cls, lab_color_vector, lab_color_matrix, Kl=1, Kc=1, Kh=1):
        return tf.math.pow(cls.ciede2000_tensor(lab_color_vector, lab_color_matrix, Kl=Kl, Kc=Kc, Kh=Kh, loss_mode=3, epsilon=1e-16), 3)


class LossFunc:
    def get_loss_func(self):
        raise NotImplementedError()  # make it virtual/abstract
    def get_loss_func_name(self):
        raise NotImplementedError()  # make it virtual/abstract


def my_print(dbg, *args):
    if dbg:
        print("")
        print(" ".join(map(str,args)))
        print("")
    else:
        pass


# the pool of loss functions for appearance (color and/or transparency)
class AppearanceLossPool(LossFunc):
    def __init__(self, labPredicted=False, specPredicted=False, auxspecPredicted=False, alphaPredicted=False, cmykgPredicted=False,
                 lab_loss=None, spec_loss=None, auxspec_loss=None, alpha_loss=None, boundary_loss=None, allOnes_loss=None,
                 Elab_weight=1.0, Espec_weight=1.0, Eauxspec_weight=1.0, Ealpha_weight=1.0, Eboundary_weight=1.0, EallOnes_weight=1.0,
                 labScaled=False,
                 spec_ref_in_gt=False,
                 interp_ref=None,
                 cmykg_noMul=False,
                 cmykg_sigmoidAfterMul=False,
                 cmykg_clipAfterMul=False,
                 dbg=False):
        # loss switcher
        self.labPredicted = labPredicted
        self.specPredicted = specPredicted
        self.auxspecPredicted = auxspecPredicted
        self.alphaPredicted = alphaPredicted
        if self.auxspecPredicted:
            assert(self.specPredicted)
        # interp
        self.cmykgPredicted = cmykgPredicted
        self.cmykg_noMul = cmykg_noMul
        self.cmykg_sigmoidAfterMul = cmykg_sigmoidAfterMul
        self.cmykg_clipAfterMul = cmykg_clipAfterMul
        self.interp_ref = copy.deepcopy(interp_ref)
        if self.cmykgPredicted:
            assert(self.interp_ref is not None)
        if self.interp_ref is not None:
            self.tonal_dim = len(self.interp_ref.shape) - 1
        else:
            self.tonal_dim = None
        # loss type
        self.lab_loss = lab_loss
        self.spec_loss = spec_loss
        self.auxspec_loss = auxspec_loss
        self.alpha_loss = alpha_loss
        self.boundary_loss = boundary_loss
        self.allOnes_loss = allOnes_loss
        # loss weight
        self.Elab_weight = Elab_weight
        self.Ealpha_weight = Ealpha_weight
        self.Espec_weight = Espec_weight
        self.Eauxspec_weight = Eauxspec_weight
        self.Eboundary_weight = Eboundary_weight
        self.EallOnes_weight = EallOnes_weight
        # others
        self.labScaled = labScaled
        self.spec_ref_in_gt = spec_ref_in_gt
        # print settings
        #print("Model's output: lab={}, spec={}, auxspec={}, alpha={}, cmykg={}".format(self.labPredicted, self.specPredicted, self.auxspecPredicted, self.alphaPredicted, self.cmykgPredicted))
        #print("loss type: lab={}, spec={}, auxspec={}, alpha={}, boundary={}, allOnes={}".format(self.lab_loss, self.spec_loss, self.auxspec_loss, self.alpha_loss, self.boundary_loss, self.allOnes_loss))
        #print("loss weight: lab={}, spec={}, auxspec={}, alpha={}, boundary={}, allOnes={}".format(self.Elab_weight, self.Espec_weight, self.Eauxspec_weight, self.Ealpha_weight, self.Eboundary_weight, self.EallOnes_weight))
        # dbg
        self.dbg = dbg
    def __call__(self, gt0, pred0):
        '''
        [Very Important Assumption!!!]: inside both gt & pred, data are arranged in the order of lab-spec-alpha
        '''
        # deep copy
        gt = tf.identity(gt0)
        pred = tf.identity(pred0)
        #print("__gt/pred shape:", gt.numpy().shape, pred.numpy().shape)
        #print("__gt/pred mean:", gt.numpy().mean(axis=0), pred.numpy().mean(axis=0))
        #assert((pred[:,:self.tonal_dim].numpy() >= 0.0).all()), print(pred[:,:self.tonal_dim].numpy().min())
        #assert((pred[:,:self.tonal_dim].numpy() <= 1.0).all()), print(pred[:,:self.tonal_dim].numpy().max())
        # 0. interp
        if self.cmykgPredicted:
            #print(pred.shape)
            if self.cmykg_noMul:
                CMYKG = pred[:, :self.tonal_dim]
            else:
                cmykg_pred, CMYKG = pred[:,:self.tonal_dim], pred[:,self.tonal_dim:]

                if self.dbg:
                    my_print(self.dbg, "")
                    my_print(self.dbg, "cmykg_pred min/max:", cmykg_pred.numpy().min(axis=0), cmykg_pred.numpy().max(axis=0))
                    my_print(self.dbg, "cmykg_pred mean:", cmykg_pred.numpy().mean(axis=0))
                    #cmykg_pred = tf.clip_by_value(cmykg_pred, 0.0, 1.0)
                    #my_print(self.dbg, "(after clipping) cmykg_pred min/max:", cmykg_pred.numpy().min(axis=0), cmykg_pred.numpy().max(axis=0))
                    #my_print(self.dbg, "(after clipping) cmykg_pred mean:", cmykg_pred.numpy().mean(axis=0))
                    my_print(self.dbg, "(before mutiply with cmykg) CMYKG mean:", CMYKG.numpy().mean(axis=0))
                CMYKG = tf.math.multiply(CMYKG, cmykg_pred)
                if self.cmykg_sigmoidAfterMul:
                    #print("Apply sigmoid after mul when calculating loss.")
                    CMYKG = tf.math.sigmoid(CMYKG)
                if self.cmykg_clipAfterMul:
                    #print("Clip after mul when calculating loss.")
                    tf.clip_by_value(CMYKG, 0.0, 1.0)
            if self.dbg:
                my_print(self.dbg, "(after mutiply with cmykg) CMYKG mean:", CMYKG.numpy().mean(axis=0))
            #print("self.interp_ref.shape:", self.interp_ref.shape)
            interp = tfp.math.batch_interp_regular_nd_grid(x=CMYKG,
                                                           x_ref_min=tf.zeros([self.tonal_dim], dtype=CMYKG.dtype),
                                                           x_ref_max=tf.ones([self.tonal_dim], dtype=CMYKG.dtype),
                                                           y_ref=tf.constant(self.interp_ref, dtype=CMYKG.dtype),
                                                           axis=0)
            '''
            print("!!! fake interp !!!")
            interp = tf.linalg.matmul(CMYKG, tf.ones([5,32])) / 5.
            '''
            #print("interp shape:", interp.numpy().shape)
            if self.dbg:
                my_print(self.dbg, "interp mean:", interp.shape, interp.numpy().mean(axis=0))
                my_print(self.dbg, "interp min:", interp.shape, interp.numpy().min(axis=0))
                my_print(self.dbg, "interp max:", interp.shape, interp.numpy().max(axis=0))
                my_print(self.dbg, "interp[0]:", interp[0])
                my_print(self.dbg, "")
            pred = interp
        # 1. parse outputs
        #print("gt/pred shape:", gt.numpy().shape, pred.numpy().shape)
        #print("gt/pred mean:", gt.numpy().mean(axis=0), pred.numpy().mean(axis=0))
        #assert((pred.numpy() >= 0.0).all()), print(pred.numpy().min())
        #assert((pred.numpy() <= 1.0).all()), print(pred.numpy().max())
        lab1, lab2 = None, None
        spec1, spec2 = None, None
        alpha1, alpha2 = None, None
        idx_output_start = None
        idx_output_end = 0
        if self.labPredicted:
            idx_output_start = idx_output_end
            idx_output_end += 3
            lab1, lab2 = gt[:, idx_output_start:idx_output_end], pred[:, idx_output_start:idx_output_end]
        if self.specPredicted:
            idx_output_start = idx_output_end
            idx_output_end += 31
            spec1, spec2 = gt[:, idx_output_start:idx_output_end], pred[:, idx_output_start:idx_output_end]
            if self.spec_ref_in_gt:
                gt_shape1 = tf.shape(gt)[1]
                spec_ref = gt[:, gt_shape1//2+idx_output_start:gt_shape1//2+idx_output_end]
                spec2 += spec_ref
                spec1 += spec_ref
            if not self.labPredicted and self.lab_loss is not None:
                lab1, lab2 = LABfromSpectra_tf(spec1), LABfromSpectra_tf(spec2)
        if lab1 is not None and self.labScaled:
            lab1 = scaleUpLab_tf(lab1)  #  scale up before calculating ciede color diff
            lab2 = scaleUpLab_tf(lab2)
        if self.auxspecPredicted:
            idx_output_start = idx_output_end
            idx_output_end += 10
            auxspec1, auxspec2 = gt[:, idx_output_start:idx_output_end], pred[:, idx_output_start:idx_output_end]
        if self.alphaPredicted:
            alpha1, alpha2 = gt[:, -1], pred[:, -1]
        assert(idx_output_start is not None) # output was parsed wrongly
        # 2. calculate individual loss
        Elab, Espec, Eauxspec, Ealpha, Eboundary, EallOnes = 0, 0, 0, 0, 0, 0
        if self.lab_loss is not None:
            assert(lab1 is not None and lab2 is not None)
            Elab = self.lab_loss(lab1, lab2)
        if self.spec_loss is not None:
            assert(spec1 is not None and spec2 is not None)
            Espec = self.spec_loss(spec1, spec2)
        if self.auxspec_loss is not None:
            assert(auxspec1 is not None and auxspec2 is not None)
            Eauxspec = self.auxspec_loss(auxspec1, auxspec2)
        if self.alpha_loss is not None:
            assert(alpha1 is not None and alpha2 is not None)
            Ealpha = self.alpha_loss(alpha1, alpha2)
        if self.boundary_loss is not None:
            Eboundary = self.boundary_loss(pred)
        if self.allOnes_loss is not None:
            EallOnes = self.allOnes_loss(pred)
        # 3. combined losses
        E_all = self.Elab_weight * Elab + \
                self.Ealpha_weight * Ealpha + \
                self.Espec_weight * Espec + \
                self.Eauxspec_weight * Eauxspec + \
                self.Eboundary_weight * Eboundary +\
                self.EallOnes_weight * EallOnes
        #my_print(self.dbg, "losses shape:", E_all.shape, Elab.shape, Espec.shape, Ealpha.numpy().shape)
        #my_print(self.dbg, "losses:", E_all.numpy().mean(), Elab.numpy().mean(), Espec.numpy().mean(), Ealpha.numpy().mean())
        #my_print(self.dbg, "losses (all/Elab/Espec):", E_all.numpy().mean(), Elab.numpy().mean(), Espec.numpy().mean())
        return E_all
    def get_loss_func(self):
        return self.__call__
    def get_loss_func_name(self):
        def trim(s):
            if s.endswith('_tensor'):
                return s[:-len('_tensor')]
            else:
                return s
        func_name_tmp = []
        if self.lab_loss is not None:
            func_name_tmp.append("lab{}{}".format(trim(self.lab_loss.__name__), self.Elab_weight))
        if self.spec_loss is not None:
            func_name_tmp.append("spec{}{}".format(trim(self.spec_loss.__name__), self.Espec_weight))
        if self.auxspec_loss is not None:
            func_name_tmp.append("auxspec{}{}".format(trim(self.auxspec_loss.__name__), self.Eauxspec_weight))
        if self.alpha_loss is not None:
            func_name_tmp.append("alpha{}{}".format(trim(self.alpha_loss.__name__), self.Ealpha_weight))
        if self.boundary_loss is not None:
            func_name_tmp.append("{}{}".format(trim(self.boundary_loss.__name__), self.Eboundary_weight))
        if self.allOnes_loss is not None:
            func_name_tmp.append("{}{}".format(trim(self.allOnes_loss.__name__), self.EallOnes_weight))
        func_name = "_".join(func_name_tmp)
        return func_name


class DeltaAlphaSpecMSE(LossFunc):
    def __init__(self, Ealpha_weight=1.0, Espec_weight=1.0):
        self.Ealpha_weight = Ealpha_weight
        self.Espec_weight = Espec_weight
    def deltaAlpha_specMSE(self, spec_and_alpha_arr1, spec_and_alpha_arr2):
        print('deltaAlpha_specRMSE: Ealpha_weight, Espec_weight =', self.Ealpha_weight, self.Espec_weight)
        spec1, alpha1 = spec_and_alpha_arr1[:, :31], spec_and_alpha_arr1[:, 31]
        spec2, alpha2 = spec_and_alpha_arr2[:, :31], spec_and_alpha_arr2[:, 31]
        Ealpha = deltaAlpha(alpha1, alpha2)
        Espec = K.mean(K.square(spec1 - spec2), axis=-1)  # MSE
        E_all = self.Ealpha_weight * Ealpha + self.Espec_weight * Espec
        return E_all
    def get_loss_func(self):
        return self.deltaAlpha_specMSE
    def get_loss_func_name(self):
        return "deltaAlpha{}_specMSE{}".format(self.Ealpha_weight, self.Espec_weight)


class Ciede2000SquareDeltaAlphaSpecRMSE(LossFunc):
    def __init__(self, Elab_weight=1.0, Ealpha_weight=1.0, Espec_weight=1.0):
        self.Elab_weight = Elab_weight
        self.Ealpha_weight = Ealpha_weight
        self.Espec_weight = Espec_weight
    def ciede2000Square_deltaAlpha_specRMSE(self, spec_and_alpha_arr1, spec_and_alpha_arr2, Kl=1, Kc=1, Kh=1):
        print('ciede2000Square_deltaAlpha_specRMSE: Elab_weight, Ealpha_weight, Espec_weight =',
              self.Elab_weight, self.Ealpha_weight, self.Espec_weight)
        spec1, alpha1 = spec_and_alpha_arr1[:, :31], spec_and_alpha_arr1[:, 31]
        spec2, alpha2 = spec_and_alpha_arr2[:, :31], spec_and_alpha_arr2[:, 31]
        lab_arr1 = LABfromSpectra_tf(spec1)
        lab_arr2 = LABfromSpectra_tf(spec2)
        Elab = my_delta_e_cie2000_noMean(lab_arr1, lab_arr2, Kl=Kl, Kc=Kc, Kh=Kh)
        Elab = K.square(Elab)  # square on ciede2000
        Ealpha = deltaAlpha(alpha1, alpha2)
        Espec = K.sqrt(K.mean(K.square(spec1 - spec2), axis=-1))
        E_all = self.Elab_weight * Elab + self.Ealpha_weight * Ealpha + self.Espec_weight * Espec
        return E_all
    def get_loss_func(self):
        return self.ciede2000Square_deltaAlpha_specRMSE
    def get_loss_func_name(self):
        return "ciede2000Square{}_deltaAlpha{}_specRMSE{}".format(self.Elab_weight, self.Ealpha_weight, self.Espec_weight)


class Ciede2000DeltaAlphaSpecRMSE(LossFunc):
    def __init__(self, Elab_weight=1.0, Ealpha_weight=1.0, Espec_weight=1.0):
        self.Elab_weight = Elab_weight
        self.Ealpha_weight = Ealpha_weight
        self.Espec_weight = Espec_weight
    def ciede2000_deltaAlpha_specRMSE(self, spec_and_alpha_arr1, spec_and_alpha_arr2, Kl=1, Kc=1, Kh=1):
        print('ciede2000_deltaAlpha_specRMSE: Elab_weight, Ealpha_weight, Espec_weight =',
              self.Elab_weight, self.Ealpha_weight, self.Espec_weight)
        spec1, alpha1 = spec_and_alpha_arr1[:, :31], spec_and_alpha_arr1[:, 31]
        spec2, alpha2 = spec_and_alpha_arr2[:, :31], spec_and_alpha_arr2[:, 31]
        lab_arr1 = LABfromSpectra_tf(spec1)
        lab_arr2 = LABfromSpectra_tf(spec2)
        Elab = my_delta_e_cie2000_noMean(lab_arr1, lab_arr2, Kl=Kl, Kc=Kc, Kh=Kh)
        Ealpha = deltaAlpha(alpha1, alpha2)
        Espec = K.sqrt(K.mean(K.square(spec1 - spec2), axis=-1))
        E_all = self.Elab_weight * Elab + self.Ealpha_weight * Ealpha + self.Espec_weight * Espec
        return E_all
    def get_loss_func(self):
        return self.ciede2000_deltaAlpha_specRMSE
    def get_loss_func_name(self):
        return "ciede2000{}_deltaAlpha{}_specRMSE{}".format(self.Elab_weight, self.Ealpha_weight, self.Espec_weight)


class Ciede2000DeltaAlphaSpecMSE(LossFunc):
    def __init__(self, Elab_weight=1.0, Ealpha_weight=1.0, Espec_weight=1.0):
        self.Elab_weight = Elab_weight
        self.Ealpha_weight = Ealpha_weight
        self.Espec_weight = Espec_weight
    def ciede2000_deltaAlpha_specMSE(self, spec_and_alpha_arr1, spec_and_alpha_arr2, Kl=1, Kc=1, Kh=1):
        print('ciede2000_deltaAlpha_specMSE: Elab_weight, Ealpha_weight, Espec_weight =',
              self.Elab_weight, self.Ealpha_weight, self.Espec_weight)
        spec1, alpha1 = spec_and_alpha_arr1[:, :31], spec_and_alpha_arr1[:, 31]
        spec2, alpha2 = spec_and_alpha_arr2[:, :31], spec_and_alpha_arr2[:, 31]
        lab_arr1 = LABfromSpectra_tf(spec1)
        lab_arr2 = LABfromSpectra_tf(spec2)
        Elab = my_delta_e_cie2000_noMean(lab_arr1, lab_arr2, Kl=Kl, Kc=Kc, Kh=Kh)
        Ealpha = deltaAlpha(alpha1, alpha2)
        Espec = K.mean(K.square(spec1 - spec2), axis=-1)
        E_all = self.Elab_weight * Elab + self.Ealpha_weight * Ealpha + self.Espec_weight * Espec
        return E_all
    def get_loss_func(self):
        return self.ciede2000_deltaAlpha_specRMSE
    def get_loss_func_name(self):
        return "ciede2000{}_deltaAlpha{}_specMSE{}".format(self.Elab_weight, self.Ealpha_weight, self.Espec_weight)


class Ciede2000DeltaAlphaSpecED(LossFunc):
    def __init__(self, Elab_weight=1.0, Ealpha_weight=1.0, Espec_weight=1.0):
        self.Elab_weight = Elab_weight
        self.Ealpha_weight = Ealpha_weight
        self.Espec_weight = Espec_weight
    def ciede2000_deltaAlpha_specED(self, spec_and_alpha_arr1, spec_and_alpha_arr2, Kl=1, Kc=1, Kh=1):
        print('ciede2000_deltaAlpha_specED: Elab_weight, Ealpha_weight, Espec_weight =',
              self.Elab_weight, self.Ealpha_weight, self.Espec_weight)
        spec1, alpha1 = spec_and_alpha_arr1[:, :31], spec_and_alpha_arr1[:, 31]
        spec2, alpha2 = spec_and_alpha_arr2[:, :31], spec_and_alpha_arr2[:, 31]
        lab_arr1 = LABfromSpectra_tf(
            spec1) * 0 + 1e-8  # add 1e-8 to stablilize the training, otherwise the loss could be inf or nan when lr is big
        lab_arr2 = LABfromSpectra_tf(spec2) * 0 + 1e-8
        Elab = my_delta_e_cie2000_noMean(lab_arr1, lab_arr2, Kl=Kl, Kc=Kc, Kh=Kh)
        Ealpha = deltaAlpha(alpha1, alpha2)
        Espec = euclidean_distance(spec1, spec2)  # K.mean(K.square(spec1 - spec2), axis=-1)
        E_all = self.Elab_weight * Elab + self.Ealpha_weight * Ealpha + self.Espec_weight * Espec
        return E_all
    def get_loss_func(self):
        return self.ciede2000_deltaAlpha_specRMSE
    def get_loss_func_name(self):
        return "ciede2000{}_deltaAlpha{}_specED{}".format(self.Elab_weight, self.Ealpha_weight, self.Espec_weight)


def euclidean_distance(y_true, y_pred):
    #print('==>euclidean_distance')
    return K.sqrt(K.sum(K.square(y_pred - y_true), axis=-1))


def cie76(y_true, y_pred):
    print('==>cie76')
    return K.sqrt(K.sum(K.square(y_pred - y_true), axis=-1))


def square_cie76(y_true, y_pred):
    print('==>square_cie76')
    return K.sum(K.square(y_pred - y_true), axis=-1)


def cie76_scalar(y_true, y_pred):
    return np.sqrt(np.sum(np.square(y_pred - y_true), axis=-1))


def cie94(y_true, y_pred):
    print('==>cie94')
    d_L = y_true[:, 0] - y_pred[:, 0]
    d_a = y_true[:, 1] - y_pred[:, 1]
    d_b = y_true[:, 2] - y_pred[:, 2]
    C1 = K.sqrt(K.sum(K.square(y_true[:, 1:]), axis=-1))
    C2 = K.sqrt(K.sum(K.square(y_pred[:, 1:]), axis=-1))
    d_C = C1 - C2
    d_H = K.sqrt(K.square(d_a) + K.square(d_b) - K.square(d_C))
    S_L = 1
    K1 = 0.045  # graphic arts
    K2 = 0.015  # graphic arts
    # K1 = 0.048 # textiles
    # K2 = 0.014 # textiles
    S_C = 1 + K1 * C1
    S_H = 1 + K2 * C1
    K_L = 1  # default
    # K_L = 2 # textiles
    K_C = 1
    K_H = 1
    d_E = K.sqrt(K.square(d_L / (K_L * S_L)) + K.square(d_C / (K_C * S_C)) - K.square(d_H / (K_H * S_H)))
    return d_E


def cie94_np(y_true, y_pred):
    print('==>cie94_np')
    d_L = y_true[:, 0] - y_pred[:, 0]
    d_a = y_true[:, 1] - y_pred[:, 1]
    d_b = y_true[:, 2] - y_pred[:, 2]
    C1 = np.sqrt(np.sum(np.square(y_true[:, 1:]), axis=-1))
    C2 = np.sqrt(np.sum(np.square(y_pred[:, 1:]), axis=-1))
    d_C = C1 - C2
    d_H = np.sqrt(np.square(d_a) + np.square(d_b) - np.square(d_C))
    S_L = 1
    K1 = 0.045  # graphic arts
    K2 = 0.015  # graphic arts
    # K1 = 0.048 # textiles
    # K2 = 0.014 # textiles
    S_C = 1 + K1 * C1
    S_H = 1 + K2 * C1
    K_L = 1  # default
    # K_L = 2 # textiles
    K_C = 1
    K_H = 1
    d_E = np.sqrt(np.square(d_L / (K_L * S_L)) + np.square(d_C / (K_C * S_C)) - np.square(d_H / (K_H * S_H)))
    return d_E


# 'mse',  'mae', 'norm_mse', 'cie76', 'square_cie76'
def my_loss_fn(y_true, y_pred, **kwargs):
    loss = kwargs['loss']
    y_true = K.cast(y_true, y_pred.dtype)
    if loss == 'mse':
        print('==>mse')
        return K.mean(K.square(y_pred - y_true), axis=-1)
    elif loss == 'mae':
        return K.mean(K.abs(y_pred - y_true), axis=-1)
    elif loss == 'norm_mse':
        norm = K.clip(K.square(y_true), K.epsilon(), None)
        # norm = K.abs(K.concatenate((y_true[:,0:1], y_true[:,0:1], y_true[:,0:1]), axis=1)) + 1e-5
        # norm = K.sqrt(K.concatenate((y_true[:,0:1], y_true[:,0:1], y_true[:,0:1]), axis=1) + 1e-8) + 1e-5
        res = K.mean(K.square(y_pred - y_true) / norm, axis=-1)
        return res
    elif loss == 'cie76':
        print('==>cie76')
        return K.sqrt(K.sum(K.square(y_pred - y_true), axis=-1))
    elif loss == 'square_cie76':
        return K.sum(K.square(y_pred - y_true), axis=-1)
    else:
        sys.exit('Error: loss not supported!')


def ED(x1,x2):
    assert(len(x1.shape) <= 2 and len(x2.shape) <= 2)
    return np.sqrt(np.sum(np.square(x1 - x2), axis=1))

def MSE(y, y_pred):
    return np.mean((y - y_pred) ** 2)

def RMS(y, y_pred):
    return np.sqrt(np.mean((y - y_pred) ** 2))

def show_ED_fn(gt, pred, percentiles=[0, 50, 90, 95, 99, 100]):
    ed = ED(gt, pred)
    print("Euclidean-Distance mean={:.4}, std={:.4}, percentiles {}={}".format(
        np.mean(ed), np.std(ed), str(percentiles), np.percentile(ed, percentiles)))

def add_dict(dict_dst, dict_src):
    for key, val in dict_src.items():
        if key in dict_dst:
            print(dict_src)
            print(dict_dst)
            sys.exit("key '{}' already exists in dict_dst!")
        else:
            dict_dst[key] = val

def cmykg_to_final_pred(pred, interp_points, interp_values, data_content='cmykg_noMul', cmykg_sigmoidAfterMul=False, cmykg_clipAfterMul=False):
    pred = copy.deepcopy(pred)
    tonal_dim = len(interp_values.shape) - 1

    if data_content == "cmykg_noMul":
        CMYKG = pred[:, :tonal_dim]
    else:
        cmykg_pred, CMYKG = pred[:, :tonal_dim], pred[:, tonal_dim:]
        #print("before mul, cmykg:\n", cmykg_pred[0:3])
        #print("before mul, CMYKG:\n", CMYKG[0:3])
        CMYKG = np.multiply(CMYKG, cmykg_pred)
        #print("after mul, CMYKG:\n", CMYKG[0:3])
        if cmykg_sigmoidAfterMul:
            print("Apply sigmoid after mul when getting metrics.")
            CMYKG = sigmoid(CMYKG)
        if cmykg_clipAfterMul:
            #print("Clip after mul when getting metrics.")
            CMYKG = np.clip(CMYKG, 0.0, 1.0)
        #print("after sig&clip, CMYKG:\n", CMYKG[0:3])
    #print(np.shape(interp_points), interp_values.shape, CMYKG.shape)
    interp = interpolate.interpn(interp_points, interp_values, CMYKG, method='linear', bounds_error=True,
                                 fill_value=np.NaN)
    pred = interp

    if pred.shape[1] == 3:
        data_content = "lab"
    elif pred.shape[1] == 3 + 1:
        data_content = "labAndAlpha"
    elif pred.shape[1] == 31:
        data_content = "spec"
    elif pred.shape[1] == 31 + 1:
        data_content = "specAndAlpha"
    elif pred.shape[1] == 41:
        data_content = "fullspec"
    elif pred.shape[1] == 41 + 1:
        data_content = "fullspecAndAlpha"
    else:
        data_content = None

    return pred, data_content


def get_metric_dict(pred, gt, data_content='specAndAlpha', toPrint=True, title="result",
                    cmykg_noMul=False, cmykg_sigmoidAfterMul=False, cmykg_clipAfterMul=False,
                    interp_points=None, interp_values=None):
    if data_content == "cmykg" or data_content == "cmykg_noMul":
        tonal_dim = len(interp_values.shape) - 1
        #print("tonal_dim:", tonal_dim)
        # print("before interp, pred shape/mean/max/min:", pred.shape, pred.mean(), pred.max(), pred.min())
        pred, data_content = cmykg_to_final_pred(pred, interp_points, interp_values, data_content, cmykg_sigmoidAfterMul, cmykg_clipAfterMul)
        '''
        if data_content == "cmykg_noMul":
            CMYKG = pred[:, :tonal_dim]
        else:
            cmykg_pred, CMYKG = pred[:, :tonal_dim], pred[:, tonal_dim:]
            print("before mul, cmykg:\n", cmykg_pred[0:3])
            print("before mul, CMYKG:\n", CMYKG[0:3])
            CMYKG = np.multiply(CMYKG, cmykg_pred)
            print("after mul, CMYKG:\n", CMYKG[0:3])
            if cmykg_sigmoidAfterMul:
                print("Apply sigmoid after mul when getting metrics.")
                CMYKG = sigmoid(CMYKG)
            if cmykg_clipAfterMul:
                print("Clip after mul when getting metrics.")
                CMYKG = np.clip(CMYKG, 0.0, 1.0)
            print("after sig&clip, CMYKG:\n", CMYKG[0:3])
        print(np.shape(interp_points), interp_values.shape, CMYKG.shape)
        interp = interpolate.interpn(interp_points, interp_values, CMYKG, method='linear', bounds_error=True,
                                     fill_value=np.NaN)
        pred = interp
        '''
        #print("after interp, pred shape/mean/max/min:", pred.shape, pred.mean(), pred.max(), pred.min())

        '''
        if pred.shape[1] == 3:
            data_content = "lab"
        elif pred.shape[1] == 3+1:
            data_content = "labAndAlpha"
        elif pred.shape[1] == 31:
            data_content = "spec"
        elif pred.shape[1] == 31+1:
            data_content = "specAndAlpha"
        elif pred.shape[1] == 41:
            data_content = "fullspec"
        elif pred.shape[1] == 41+1:
            data_content = "fullspecAndAlpha"
        else:
            data_content = None
        '''
        #print("Change data_content to", data_content)
        #print("Continue getting metrics...")


    if data_content == "labAndSpecAndAlpha" or data_content == "labScaledAndSpecAndAlpha":
        lab_pred = pred[:, 0:3]
        spectral_pred = pred[:, 3:3+31]
        transl_pred = pred[:, 3+31:]
        lab_gt = gt[:, 0:3]
        spectral_gt = gt[:, 3:3+31]
        transl_gt = gt[:, 3+31:]
        # Elab: ciede2000
        Elab = color_diff(lab_pred, lab_gt, type='cie2000', Kl=1, Kc=1, Kh=1)
        Elab_stat_dict = get_statistics(Elab, percentiles=None, prefix='Elab_', toPrint=toPrint, title=title + " Elab")
        # Espec: RMSE
        Espec = RMSE(spectral_pred, spectral_gt)
        Espec_stat_dict = get_statistics(Espec, percentiles=None, prefix='_Espec_', toPrint=toPrint,
                                        title=title + " Espec")
        # Ealpha: abs error
        Ealpha = abs_err(transl_pred, transl_gt)
        Ealpha_stat_dict = get_statistics(Ealpha, percentiles=None, prefix='__Ealpha_', toPrint=toPrint,
                                         title=title + " Ealpha")
        # combined
        res = {}
        add_dict(res, Elab_stat_dict)
        add_dict(res, Espec_stat_dict)
        add_dict(res, Ealpha_stat_dict)
        return res
    elif data_content == "specAndAlpha" or data_content == "fullspecAndAlpha":
        transl_pred = pred[:, -1:]
        spectral_pred = pred[:, :31]
        if data_content == "fullspecAndAlpha":
            auxspectral_pred = pred[:, 31:41]
            auxspectral_gt = gt[:, 31:41]
        transl_gt = gt[:, -1:]
        spectral_gt = gt[:, :31]
        LAB_from_spectral_pred = get_lab_from_spectral(spectral_pred)
        LAB_from_spectral_gt = get_lab_from_spectral(spectral_gt)
        # Elab: ciede2000
        Elab = color_diff(LAB_from_spectral_pred, LAB_from_spectral_gt, type='cie2000', Kl=1, Kc=1, Kh=1)
        Elab_stat_dict = get_statistics(Elab, percentiles=None, prefix='Elab_', toPrint=toPrint, title=title + " Elab")
        # Espec: RMSE
        Espec = RMSE(spectral_pred, spectral_gt)
        Espec_stat_dict = get_statistics(Espec, percentiles=None, prefix='_Espec_', toPrint=toPrint, title=title + " Espec")
        '''
        print("=" * 100)
        print(spectral_pred.shape)
        show_ED_fn(spectral_pred, spectral_gt)
        show_ED_fn(LAB_from_spectral_pred, LAB_from_spectral_gt)
        '''
        # Eauxspec: RMSE
        if data_content == "fullspec":
            Eauxspec = RMSE(auxspectral_pred, auxspectral_gt)
            Eauxspec_stat_dict = get_statistics(Eauxspec, percentiles=None, prefix='_Eauxspec_', toPrint=toPrint, title=title + " Eauxspec")
        # Ealpha: abs error
        Ealpha = abs_err(transl_pred, transl_gt)
        Ealpha_stat_dict = get_statistics(Ealpha, percentiles=None, prefix='__Ealpha_', toPrint=toPrint, title=title + " Ealpha")
        # combined
        res = {}
        add_dict(res, Elab_stat_dict)
        add_dict(res, Espec_stat_dict)
        if data_content == "fullspec":
            add_dict(res, Eauxspec_stat_dict)
        add_dict(res, Ealpha_stat_dict)
        return res
    elif data_content == "spec" or data_content == "specOffset" or data_content == "specOffsetAndRef" or data_content == "fullspec":
        spectral_pred = pred[:, :31]
        spectral_gt = gt[:, :31]
        LAB_from_spectral_pred = get_lab_from_spectral(spectral_pred)
        LAB_from_spectral_gt = get_lab_from_spectral(spectral_gt)
        if data_content == "fullspec":
            auxspectral_pred = pred[:, 31:41]
            auxspectral_gt = gt[:, 31:41]
        # Elab: ciede2000
        Elab = color_diff(LAB_from_spectral_pred, LAB_from_spectral_gt, type='cie2000', Kl=1, Kc=1, Kh=1)
        Elab_stat_dict = get_statistics(Elab, percentiles=None, prefix='Elab_', toPrint=toPrint, title=title + " Elab")
        # Espec: RMSE
        Espec = RMSE(spectral_pred, spectral_gt)
        Espec_stat_dict = get_statistics(Espec, percentiles=None, prefix='_Espec_', toPrint=toPrint, title=title + " Espec")
        # Eauxspec: RMSE
        if data_content == "fullspec":
            Eauxspec = RMSE(auxspectral_pred, auxspectral_gt)
            Eauxspec_stat_dict = get_statistics(Eauxspec, percentiles=None, prefix='_Eauxspec_', toPrint=toPrint, title=title + " Eauxspec")
        # combined
        res = {}
        add_dict(res, Elab_stat_dict)
        add_dict(res, Espec_stat_dict)
        if data_content == "fullspec":
            add_dict(res, Eauxspec_stat_dict)
        return res
    elif data_content == "labAndAlpha" or data_content == "labScaledAndAlpha":
        transl_pred = pred[:, 3:]
        lab_pred = pred[:, :3]
        transl_gt = gt[:, 3:]
        lab_gt = gt[:, :3]
        # Elab: ciede2000
        Elab = color_diff(lab_pred, lab_gt, type='cie2000', Kl=1, Kc=1, Kh=1)
        Elab_stat_dict = get_statistics(Elab, percentiles=None, prefix='Elab_', toPrint=toPrint, title=title + " Elab")
        # Ealpha: abs error
        Ealpha = abs_err(transl_pred, transl_gt)
        Ealpha_stat_dict = get_statistics(Ealpha, percentiles=None, prefix='__Ealpha_', toPrint=toPrint, title=title + " Ealpha")
        # combined
        res = {}
        add_dict(res, Elab_stat_dict)
        add_dict(res, Ealpha_stat_dict)
        return res
    elif data_content == "lab" or data_content == "labOnly" or data_content == "labScaled" or data_content == "labOffset":
        # Elab: ciede2000
        Elab = color_diff(pred, gt, type='cie2000', Kl=1, Kc=1, Kh=1)
        Elab_stat_dict = get_statistics(Elab, percentiles=None, prefix='Elab_', toPrint=toPrint, title=title + " Elab")
        # combined
        res = {}
        add_dict(res, Elab_stat_dict)
        return res
    elif data_content == "alpha":
        transl_pred = pred[:, -1:]
        transl_gt = gt[:, -1:]
        # Ealpha: abs error
        Ealpha = abs_err(transl_pred, transl_gt)
        Ealpha_stat_dict = get_statistics(Ealpha, percentiles=None, prefix='__Ealpha_', toPrint=toPrint, title=title + " Ealpha")
        # combined
        res = {}
        add_dict(res, Ealpha_stat_dict)
        return res
    else:
        print("Not supported raw_data")
        return {}

def plot_E_lab(gt, pred, percentiles=[50,70,75,80,90], plot=True, toPrint=True, title='deltaE2000', returnElabDict=False):
    E_lab = color_diff(gt, pred, type='cie2000', Kl=1, Kc=1, Kh=1)
    mean = np.mean(E_lab)
    std = np.std(E_lab)
    median = np.median(E_lab)
    p90 = np.percentile(E_lab, 90)
    p95 = np.percentile(E_lab, 95)
    max = np.max(E_lab)
    v_percentiles = np.around(np.percentile(E_lab, percentiles), decimals=4) if percentiles is not None else None
    E_lab_dict = {"Elab_mean": mean, "Elab_std": std, "Elab_median": median,
                  "Elab_p90": p90, "Elab_p95": p95, "Elab_max": max,
                  "Elab_percentiles": percentiles, "Elab_v_percentiles": v_percentiles}
    if toPrint:
        print("gt.shape=", gt.shape)
        print("{}: mean={:.4}, std={:.4}, median={:.4}, percentiles {}={}, p90={:.4}, p95={:.4}, max={:.4}".format(
            title, mean, median, std, str(percentiles), v_percentiles, p90, p95, max))
    if plot:
        plt.figure(figsize=[15, 4])
        plt.suptitle(title)

        plt.subplot(1,2,1)
        plt.hist(E_lab)
        plt.xlabel('E_lab')
        plt.ylabel('freq.')

        plt.subplot(1,2,2)
        plt.hist(E_lab, bins=100, cumulative=True, density=1, histtype='step')
        plt.xlabel('E_lab')
        plt.ylabel('cum freq.')
        plt.plot([0, np.percentile(E_lab, 90), np.percentile(E_lab, 90)], [0.9, 0.9, 0], '--')
        plt.yticks(ticks=np.arange(0.0, 1.1, 0.1))
        plt.show()
    if returnElabDict:
        return E_lab_dict

def summarize_metric_dict_list(metric_dict_list, toPrint=True, title="deltaE2000", returnSummarizedDict=False):
    metric_dict_combined = {}
    # init
    for key in metric_dict_list[0].keys():
        metric_dict_combined[key] = []
        metric_dict_combined["meanOf_" + key] = None
        metric_dict_combined["stdOf_" + key] = None
    # put together
    for E_lab_dict_tmp in metric_dict_list:
        for key, val in E_lab_dict_tmp.items():
            if key is not "Elab_percentiles":  # type(val) is not str:
                metric_dict_combined[key].append(val)
    # analyze
    print("{}:".format(title))
    for key, val in metric_dict_combined.items():
        if key.startswith("meanOf_") or key.startswith("stdOf_") or None in val:
            continue
        metric_dict_combined["meanOf_" + key] = np.mean(val)
        metric_dict_combined["stdOf_" + key] = np.std(val)
        if toPrint:
            if len(metric_dict_list) == 1:
                print("    {} = {:.4f}".format(key, np.mean(val)))
            else:
                print("    {} mean={:.4f}, std={:.4f}".format(key, np.mean(val), np.std(val)))
    if returnSummarizedDict:
        return metric_dict_combined

def abs_err(gt, pred):
    assert(pred.shape == gt.shape)
    err = np.abs(gt - pred)
    return err

def plot_abs_err(gt, pred, percentiles=[50,70,75,80,90], plot=True):
    assert(pred.shape == gt.shape)
    err = np.abs(gt - pred)
    show_statistics(err, percentiles)
    if plot:
        plt.figure(figsize=[15, 4])
        plt.subplot(1,2,1)
        plt.hist(err)
        plt.xlabel('abs_err')
        plt.ylabel('freq.')

        plt.subplot(1,2,2)
        plt.hist(err, bins=100, cumulative=True, density=1, histtype='step')
        plt.xlabel('abs_err')
        plt.ylabel('cum freq.')
        plt.plot([0, np.percentile(err, 90), np.percentile(err, 90)], [0.9, 0.9, 0], '--')
        plt.yticks(ticks=np.arange(0.0, 1.1, 0.1))
        plt.show()

def get_statistics(err, percentiles=None, prefix='', toPrint=False, title="result", returnDict=True):
    mean = np.mean(err)
    std = np.std(err)
    median = np.median(err)
    p90 = np.percentile(err, 90)
    p95 = np.percentile(err, 95)
    p99 = np.percentile(err, 99)
    p99_9 = np.percentile(err, 99.9)
    max = np.max(err)
    v_percentiles = np.around(np.percentile(err, percentiles), decimals=4) if percentiles is not None else None
    res_dict0 = {"mean": mean, "std": std, "median": median,
                  "p90": p90, "p95": p95, "p99": p99, "p99.9": p99_9, "max": max,
                  "percentiles": percentiles, "v_percentiles": v_percentiles}
    res_dict = {}
    for key, val in res_dict0.items():
        res_dict[prefix + key] = val  # replace the key
    if toPrint:
        print(title, ":")
        if percentiles is not None:
            print("  mean={:.4}, std={:.4}, median={:.4}, p90={:.4}, p95={:.4}, p99={:.4}, p99.9={:.4}, max={:.4}, percentiles {}={}".format(
                mean, std, median, p90, p95, p99, p99_9, max, str(percentiles), v_percentiles))
        else:
            print("  mean={:.4}, std={:.4}, median={:.4}, p90={:.4}, p95={:.4}, p99={:.4}, p99.9={:.4}, max={:.4}".format(
                mean, std, median, p90, p95, p99, p99_9, max))
    if returnDict:
        return res_dict

def show_statistics(err, percentiles):
    print("mean={:.4}, std={:.4}, median={:.4}, percentiles {}={}".format( np.mean(err), np.std(err), np.median(err), str(percentiles), np.percentile(err, percentiles)))


def plot_abs_distri(a):

    a = np.abs(a)
    m = np.mean(a, axis=0)
    s = np.std(a, axis=0)
    p5 = np.percentile(a, 30, axis=0) + 1e-16
    p95 = np.percentile(a, 95, axis=0)
    plt.figure(figsize=[20,5])
    items = [m,s,p5,p95]
    names = ['mean', 'std', 'p5', 'p95']
    for i, item, name in zip(range(len(items)), items, names):
        plt.subplot(1,4,i+1)
        plt.plot(item, '.')
        plt.title(name)
    plt.show()
