import numpy as np
import pandas as pd
import copy
import time
import os

class Dataset:
    dataset_path_dict = {
        'stratasys': './data/stratasys/',
        'mimaki_1': './data/mimaki_1/',
        'mimaki_2': './data/mimaki_2/',
    }
    fname_dict = {
        'tonal': 'InputReferencePattern.txt',
        'reflectance': 'ReflectanceData.txt',
        'translucency': 'TranslucencyFile.txt',
    }
    gamma_stratified_dict = {
        'stratasys': True,
        'mimaki_1': True,
        'mimaki_2': False,
    }
    _3GP_dict = {
        'stratasys': [0, 128, 255],
        'mimaki_2': [0, 127.501946, 255]
    }
    _4GP_dict = {
        'mimaki_2': [0, 255/3, 255/3*2, 255]
    }
    tonal_dim_dict = {
        'stratasys': 5,
        'mimaki_1': 5,
        'mimaki_2': 4,
    }
    tonal_middle_point_dict = {
        'stratasys': 128,
        'mimaki_1': np.NaN,
        'mimaki_2': 127.501946
    }

    def __init__(self, dataset, scale_tonal=True, train_dedicated_samples='prim', n_test=300, seed_for_test_set=789):
        # file pathes
        self.dataset = dataset
        assert(dataset in self.dataset_path_dict), 'Only "stratasys"/"mimaki_1"/"mimaki_2" datasets are supported!'
        self.dataset_path = self.dataset_path_dict[self.dataset]
        print(f"dataset is {self.dataset}, dataset path is {self.dataset_path}")
        self.tonal_fpath = os.path.join(self.dataset_path, self.fname_dict['tonal'])
        self.refl_fpath = os.path.join(self.dataset_path, self.fname_dict['reflectance'])
        self.trans_fpath = os.path.join(self.dataset_path, self.fname_dict['translucency'])
        self.tonal_arr, self.refl_arr, self.trans_arr = None, None, None
        # tonal dim
        self.tonal_dim = self.tonal_dim_dict[self.dataset]
        # settings for scaling
        self.scale_tonal = scale_tonal
        self.tonal_lbournd, self.tonal_ubound = 0., 255.
        self.tonal_middle_point = self.tonal_middle_point_dict[self.dataset]
        if self.scale_tonal:
            self.tonal_lbournd, self.tonal_ubound = self.tonal_lbournd /255., self.tonal_ubound / 255.
            self.tonal_middle_point /= 255.
            for _dict in [self._3GP_dict, self._4GP_dict]:
                for k in _dict:
                    _dict[k] = [i/255. for i in _dict[k]]
        # settings for data split
        self.train_dedicated_samples = train_dedicated_samples
        self.n_test = n_test
        self.seed_for_test_set = seed_for_test_set
        self.train_dedicated_idx = []
        self.selectable_idx = None # not-yet-selected samples
        self.rnd_idx = None # mimaki_2 has 1099 random samples
        self.gamma_stratified_sampling = self.gamma_stratified_dict[self.dataset]
        self.hiPrio_idx_for_P = None
        self.train_idx, self.valid_idx, self.test_idx = None, None, None

    def load_data(self):
        # load tonal
        self.tonal_arr = parse_CMYKT(self.tonal_fpath, pandas=False, skiprows=3, skipfooter=1)
        self.tonal_arr = self.tonal_arr[:, :self.tonal_dim]
        if self.scale_tonal:
            self.tonal_arr /= 255.
            print(f"tonal is scaled down by 255X, leading to range [{self.tonal_arr.min()}, {self.tonal_arr.max()}]")
        # load reflectance
        self.refl_arr = parse_CMYKT(self.refl_fpath, pandas=False, skiprows=3, skipfooter=1)

        if self.dataset != 'mimaki_2':
            self.trans_arr = parse_CMYKT(self.trans_fpath, pandas=False, skiprows=3, skipfooter=1)
            self.trans_arr = self.trans_arr[:,0:1]
         # load translucency
        if self.trans_arr is not None:
            assert(self.tonal_arr.shape[0] == self.refl_arr.shape[0] == self.trans_arr.shape[0])
        else:
            assert (self.tonal_arr.shape[0] == self.refl_arr.shape[0])

        self.n_sample = self.tonal_arr.shape[0]
        print(f"Loaded {self.n_sample} samples from {self.dataset_path}")
        print(f"tonal data: shape={self.tonal_arr.shape}, range={self.tonal_arr.min(), self.tonal_arr.max()}")
        print(f"reflectance data: shape={self.refl_arr.shape}, range={self.refl_arr.min(), self.refl_arr.max()}")
        if self.trans_arr is not None:
            print(f"translucency data: shape={self.trans_arr.shape}, range={self.trans_arr.min(), self.trans_arr.max()}")

        # update selectable_idx
        self.selectable_idx = list(range(self.n_sample))

        # set rnd_idx for mimaki_2
        if self.dataset == 'mimaki_2':
            self.rnd_idx = [i for i in list(range(self.n_sample)) if i >= 2400 and i != 3400]

        self.set_train_dedicated_idx()
        self.split_test_idx()
        #return self.tonal_arr, self.refl_arr, self.trans_arr


    def get_input(self):
        return copy.deepcopy(self.tonal_arr)

    def get_output(self):
        if self.dataset == 'mimaki_2':
            return copy.deepcopy(self.refl_arr)
        else:
            refl_trans_arr = np.concatenate([self.refl_arr, self.trans_arr], axis=-1)
            return copy.deepcopy(refl_trans_arr)

    def get_data(self):
        return copy.deepcopy(self.tonal_arr), copy.deepcopy(self.refl_arr), copy.deepcopy(self.trans_arr)

    def get_prim_idx(self):
        prim_idx = []
        for i in range(self.n_sample):
            if np.logical_or(self.tonal_arr[i] == self.tonal_lbournd, self.tonal_arr[i] == self.tonal_ubound).all():
                prim_idx.append(i)
        return copy.deepcopy(prim_idx)

    def get_3GP_idx(self):
        if self.dataset not in ['stratasys', 'mimaki_2']:
            return None
        _3GP = self._3GP_dict[self.dataset]
        _3GP_idx = []
        for i in range(self.n_sample):
            if np.logical_or(
                    np.logical_or(self.tonal_arr[i] == _3GP[0], self.tonal_arr[i] == _3GP[1]),
                    self.tonal_arr[i] == _3GP[2]
            ).all():
                _3GP_idx.append(i)
        return copy.deepcopy(_3GP_idx)

    def get_4GP_idx(self):
        if self.dataset != 'mimaki_2':
            return None
        _4GP = self._4GP_dict[self.dataset]
        _4GP_idx = []
        for i in range(self.n_sample):
            if (np.array([_ in _4GP for _ in self.tonal_arr[i]]) == True).all():
                _4GP_idx.append(i)
        return copy.deepcopy(_4GP_idx)

    def get_CN_ref_points(self):
        n_grids = np.NaN
        if self.train_dedicated_samples == '3GP':
            n_grids = 3
        elif self.train_dedicated_samples == '4GP':
            n_grids = 4
        else:
            return None
        return [np.linspace(self.tonal_lbournd, self.tonal_ubound, num=n_grids)] * self.tonal_dim

    @staticmethod
    def float_equal(f1, f2, epsilon=1e-6):
        return np.fabs(f1 - f2) <= epsilon

    def get_CN_ref_values(self):
        if self.dataset not in ['stratasys', 'mimaki_2']:
            return None
        x_ref_min = np.ones([self.tonal_dim], dtype="float32") * self.tonal_lbournd
        x_ref_max = np.ones([self.tonal_dim], dtype="float32") * self.tonal_ubound
        outputs = self.get_output()
        GP_idx = self.get_3GP_idx()
        prim_4grids_idx = self.get_4GP_idx()
        if self.train_dedicated_samples == '3GP':
            n_grids = 3
            GP_idx = GP_idx
        elif self.train_dedicated_samples == '4GP':
            n_grids = 4
            GP_idx = prim_4grids_idx
        else:
            return None

        interp_ref_values = np.ndarray(shape=[n_grids] * self.tonal_dim + [self.get_output().shape[-1]])
        interp_ref_values[:] = np.NaN
        cnt = 0
        if self.tonal_dim == 4:
            c, m, y, k = np.meshgrid(
                np.linspace(x_ref_min[0], x_ref_max[0], num=n_grids),
                np.linspace(x_ref_min[1], x_ref_max[1], num=n_grids),
                np.linspace(x_ref_min[2], x_ref_max[2], num=n_grids),
                np.linspace(x_ref_min[3], x_ref_max[3], num=n_grids),
                indexing='ij')
            for i0 in range(n_grids):
                for i1 in range(n_grids):
                    for i2 in range(n_grids):
                        for i3 in range(n_grids):
                            for idx in GP_idx:
                                if self.train_dedicated_samples == '3GP':
                                    if (self.tonal_arr[idx, 0] == c[i0, i1, i2, i3] or ( c[i0, i1, i2, i3] == 0.5 and abs( self.tonal_arr[idx, 0] - self.tonal_middle_point) < 1e-5)) and \
                                        (self.tonal_arr[idx, 1] == m[i0, i1, i2, i3] or ( m[i0, i1, i2, i3] == 0.5 and abs( self.tonal_arr[idx, 1] - self.tonal_middle_point) < 1e-5)) and \
                                        (self.tonal_arr[idx, 2] == y[i0, i1, i2, i3] or ( y[i0, i1, i2, i3] == 0.5 and abs( self.tonal_arr[idx, 2] - self.tonal_middle_point) < 1e-5)) and \
                                        (self.tonal_arr[idx, 3] == k[i0, i1, i2, i3] or ( k[i0, i1, i2, i3] == 0.5 and abs( self.tonal_arr[idx, 3] - self.tonal_middle_point) < 1e-5)):
                                        interp_ref_values[i0, i1, i2, i3] = outputs[idx]
                                        cnt += 1
                                        break
                                elif self.train_dedicated_samples == '4GP':
                                    if  self.tonal_arr[idx, 0] == c[i0, i1, i2, i3] and \
                                        self.tonal_arr[idx, 1] == m[i0, i1, i2, i3] and \
                                        self.tonal_arr[idx, 2] == y[i0, i1, i2, i3] and \
                                        self.tonal_arr[idx, 3] == k[i0, i1, i2, i3]:
                                        interp_ref_values[i0, i1, i2, i3] = outputs[idx]
                                        cnt += 1
                                        break
                                else:
                                    return None
        elif self.tonal_dim == 5:
            c, m, y, k, g = np.meshgrid(
                np.linspace(x_ref_min[0], x_ref_max[0], num=n_grids),
                np.linspace(x_ref_min[1], x_ref_max[1], num=n_grids),
                np.linspace(x_ref_min[2], x_ref_max[2], num=n_grids),
                np.linspace(x_ref_min[3], x_ref_max[3], num=n_grids),
                np.linspace(x_ref_min[4], x_ref_max[4], num=n_grids),
                indexing='ij')
            for i0 in range(n_grids):
                for i1 in range(n_grids):
                    for i2 in range(n_grids):
                        for i3 in range(n_grids):
                            for i4 in range(n_grids):
                                for idx in GP_idx:
                                    if self.train_dedicated_samples == '3GP':
                                        if  (self.tonal_arr[idx, 0] == c[i0, i1, i2, i3, i4] or ( c[i0, i1, i2, i3, i4] == 0.5 and abs( self.tonal_arr[idx, 0] - self.tonal_middle_point) < 1e-5)) and \
                                        (self.tonal_arr[idx, 1] == m[i0, i1, i2, i3, i4] or ( m[i0, i1, i2, i3, i4] == 0.5 and abs( self.tonal_arr[idx, 1] - self.tonal_middle_point) < 1e-5)) and \
                                        (self.tonal_arr[idx, 2] == y[i0, i1, i2, i3, i4] or ( y[i0, i1, i2, i3, i4] == 0.5 and abs( self.tonal_arr[idx, 2] - self.tonal_middle_point) < 1e-5)) and \
                                        (self.tonal_arr[idx, 3] == k[i0, i1, i2, i3, i4] or ( k[i0, i1, i2, i3, i4] == 0.5 and abs( self.tonal_arr[idx, 3] - self.tonal_middle_point) < 1e-5)) and \
                                        (self.tonal_arr[idx, 4] == g[i0, i1, i2, i3, i4] or ( g[i0, i1, i2, i3, i4] == 0.5 and abs( self.tonal_arr[idx, 4] - self.tonal_middle_point) < 1e-5)):
                                            interp_ref_values[i0, i1, i2, i3, i4] = outputs[idx]
                                            cnt += 1
                                            break
                                    elif self.train_dedicated_samples == '4GP':
                                        if  self.tonal_arr[idx, 0] == c[i0, i1, i2, i3, i4] and \
                                            self.tonal_arr[idx, 1] == m[i0, i1, i2, i3, i4] and \
                                            self.tonal_arr[idx, 2] == y[i0, i1, i2, i3, i4] and \
                                            self.tonal_arr[idx, 3] == k[i0, i1, i2, i3, i4] and \
                                            self.tonal_arr[idx, 4] == g[i0, i1, i2, i3, i4]:
                                            interp_ref_values[i0, i1, i2, i3, i4] = outputs[idx]
                                            cnt += 1
                                            break
                                    else:
                                        return None
        else:
            return None
        #print(interp_ref_values.shape, interp_ref_values.mean())
        #print("cnt:", cnt)
        assert (cnt == np.prod(interp_ref_values.shape[:-1]))
        assert (not np.isnan(interp_ref_values).any())
        return interp_ref_values

    def set_train_dedicated_idx(self):
        fn_dict = {
            'prim': self.get_prim_idx,
            '3GP': self.get_3GP_idx,
            '4GP': self.get_4GP_idx
        }
        self.train_dedicated_idx = fn_dict[self.train_dedicated_samples]()
        for i in self.train_dedicated_idx:
            self.selectable_idx.remove(i)
        print(f"Always include {self.train_dedicated_samples} samples (#={len(self.train_dedicated_idx)}) for training.")

    def get_selectable_idx(self):
        return copy.deepcopy(self.selectable_idx)

    def set_hiPrio_idx_for_P(self, idx):
        self.hiPrio_idx_for_P = idx
        print(f"Set {len(self.hiPrio_idx_for_P)} high-priority samples for P (train+valid) selection")

    def get_hiPrio_idx_for_P(self):
        return copy.deepcopy(self.hiPrio_idx_for_P)


    def split_test_idx(self):
        if self.test_idx is not None: # if test_idx has been set, then simply return it
            return copy.deepcopy(self.test_idx)

        np.random.seed(self.seed_for_test_set)
        if self.gamma_stratified_sampling:
            gammas = np.sort(np.unique(self.tonal_arr[:, -1])).tolist()
            idx_nonDedicated_stratifiedList = gamma_stratified_idx(self.selectable_idx, self.tonal_arr, gammas)
            #print(f"Before setting test set (#={self.n_test}), the remaining #samples in each gamma level:")
            for i in range(len(gammas)):
                pass #print(f"   gamma {i}, #samples={len(idx_nonDedicated_stratifiedList[i])}")
            test_idx = []
            n_added = 0
            while (n_added < self.n_test):
                for idx_nonDedicated_this_gamma in idx_nonDedicated_stratifiedList:
                    assert (len(idx_nonDedicated_this_gamma) > 0)
                    _j = np.random.choice(idx_nonDedicated_this_gamma, size=None, replace=False)
                    idx_nonDedicated_this_gamma.remove(_j)
                    test_idx.append(_j)
                    n_added += 1
                    if n_added == self.n_test:
                        break
            #print(f"After setting test set (#={self.n_test}), the remaining #samples in each gamma level:")
            for i in range(len(gammas)):
                pass #print(f"   gamma {i}, #samples={len(idx_nonDedicated_stratifiedList[i])}")
        else:
            test_idx = np.random.choice(self.rnd_idx, size=self.n_test, replace=False).tolist()

        self.test_idx = test_idx
        assert(self.n_test == len(self.test_idx))
        print(f"Selected {self.n_test} samples as test set")

        # update selectable idx
        for i in self.test_idx:
            self.selectable_idx.remove(i)
        # set hiPrio_idx_for_P
        if self.dataset == 'mimaki_2':
            self.hiPrio_idx_for_P = [i for i in self.rnd_idx if i not in self.test_idx]
        else:
            self.hiPrio_idx_for_P = self.selectable_idx

        return copy.deepcopy(self.test_idx)

# cannot import name 'ED' from 'util.cost_util'
#from util.cost_util import ED
def ED(x1,x2):
    assert(len(x1.shape) <= 2 and len(x2.shape) <= 2)
    return np.sqrt(np.sum(np.square(x1 - x2), axis=1))

def parse_CMYKT(fname, pandas=False, delimiter='\t', skiprows=0, skipfooter=0):
    if pandas:
        res = pd.read_csv(fname, dtype='float', header=None, names=['c', 'm', 'y', 'k', 't'], delimiter=delimiter,
                          skiprows=skiprows, skipfooter=skipfooter, engine='python')
    else:
        res = np.loadtxt(fname, dtype='float', delimiter=delimiter)
    return res

def parse_spectral(fname, pandas=False, delimiter='\t', skiprows=0, skipfooter=0):
    if pandas:
        res = pd.read_csv(fname, dtype='float', header=None, names=None, delimiter=delimiter,
                          skiprows=skiprows, skipfooter=skipfooter, engine='python')
    else:
        res = np.loadtxt(fname, dtype='float', delimiter=delimiter)
    return res

def parse_lab(fname, pandas=False, delimiter='\t', skiprows=0, skipfooter=0):
    if pandas:
        res = pd.read_csv(fname, dtype='float', header=None, names=['l', 'a', 'b'], delimiter=delimiter,
                          skiprows = skiprows, skipfooter = skipfooter, engine = 'python')
    else:
        res = np.loadtxt(fname, dtype='float', delimiter='\t')
    return res


def parse_alpha(fname, pandas=False, delimiter='\t', skiprows=0, skipfooter=0):
    if pandas:
        res = pd.read_csv(fname, dtype='float', header=None, names=['alpha'], delimiter=delimiter,
                          skiprows=skiprows, skipfooter=skipfooter, engine='python')
    else:
        res = np.loadtxt(fname, dtype='float', delimiter=delimiter)
    return res


def norm_features(X, degree, intercept=False, incld_orig=False):
    assert (len(X.shape) == 2)
    feats = []
    if intercept:
        feats.append(np.ones([X.shape[0], 1]))
    if incld_orig:
        feats.append(copy.deepcopy(X))

    for d in range(2, degree + 1):
        pow = 1. / d
        feat = X ** pow
        feats.append(feat)
    if len(feats) > 0:
        feats = np.concatenate(feats, axis=1)
    else:
        feats = np.array([])
    return feats

# remember the seed, then we'r able to re-do the same random splitting using the same seed
def train_test_split(X, y, ratio, ratio2=None, seed=123):
    assert(X.shape[0] == y.shape[0])
    if seed is not None:
        np.random.seed(seed)
    #print("np.random.get_state():\n", np.random.get_state())
    idx = np.random.permutation(X.shape[0])
    num_train = int(X.shape[0] * ratio)
    if ratio2 is None:
        idx_train, idx_test = idx[0:num_train], idx[num_train:]
    else:
        assert(ratio + ratio2 <= 1.0)
        num_test = int(X.shape[0] * ratio2)
        idx_train, idx_test = idx[0:num_train], idx[num_train:num_train+num_test]
    X_train, X_test = X[idx_train], X[idx_test]
    y_train, y_test = y[idx_train], y[idx_test]
    return X_train, X_test, y_train, y_test, idx_train, idx_test

def skipSomeIdx(idx, idx_to_skip):
    print("idx_to_skip[0:32]:", idx_to_skip[0:32])
    print("Before skipping idx: np.shape(idx)", np.shape(idx))
    idx = np.array([i for i in idx if i not in idx_to_skip], dtype='int32')
    print("After skipping idx: np.shape(idx)", np.shape(idx))
    return idx

def removeSomeIdx(idx, idx_to_remove):
    orig_amount = np.shape(idx)[0]
    #print("idx_to_remove[0:32]:", idx_to_remove[0:32])
    #print("Before removing idx: np.shape(idx)", np.shape(idx))
    idx = np.array([i for i in idx if i not in idx_to_remove], dtype='int32')
    #print("After removing idx: np.shape(idx)", np.shape(idx))
    return idx, orig_amount - np.shape(idx)[0]

def remove_n_idx(idx, n, except_idx=None):
    idx_to_remove = []
    for i in idx:
        if len(idx_to_remove) == n:
            break
        if except_idx is None or i not in except_idx: # safe to remove it
            idx_to_remove.append(i)
    idx = np.array([i for i in idx if i not in idx_to_remove], dtype='int32')
    return idx, np.array(idx_to_remove, dtype='int32')

def addSomeIdx(idx, idx_to_add):
    orig_amount = np.shape(idx)[0]
    #print("idx_to_add[0:32]:", idx_to_add[0:32])
    #print("Before adding idx: np.shape(idx)", np.shape(idx))
    idx_to_add_ = np.array([i for i in idx_to_add if i not in idx], dtype='int32')
    #print("idx_to_add_[0:32]:", idx_to_add_[0:32])
    idx = np.concatenate([idx, idx_to_add_])
    #print("After adding idx: np.shape(idx)", np.shape(idx))
    return idx, np.shape(idx)[0] - orig_amount

def gamma_stratified_idx(orig_idx_list, X, gamma_list):
    new_idx_list = [[] for _ in range(len(gamma_list))]
    for _idx in orig_idx_list:
        new_idx_list[gamma_list.index(X[_idx, -1])].append(_idx)
    return new_idx_list

def train_test_split_v4(X, y,
                        ratio, ratio2, ratio3=None,
                        idx_dedicated_for_training=None,
                        test_set=None,
                        idx_hiPrio_for_P=None,
                        distance_based_P=True,
                        idx_pool_for_valid=None,
                        gamma_stratified_sampling=False,
                        ensure_strata_balance=False,
                        seed=None,
                        seed_trainValidSplit=None,
                        deterministicSplitMode=1):
    if seed is not None:
        np.random.seed(seed)

    assert (X.shape[0] == y.shape[0])
    N, D = X.shape

    idx_dedicated_for_training = [] if idx_dedicated_for_training is None else copy.deepcopy(idx_dedicated_for_training)

    num_train = int(N * ratio) if ratio <= 1.0 else int(ratio)
    #print("train: #dedicated & #others:", len(idx_dedicated_for_training), num_train - len(idx_dedicated_for_training))
    assert (num_train >= len(idx_dedicated_for_training)), print("num_train={}, #dedicated={}".format(num_train, len(idx_dedicated_for_training)))
    num_valid = int(N * ratio2) if ratio2 <= 1.0 else int(ratio2)
    num_train_plus_valid = num_train + num_valid
    if idx_pool_for_valid is not None:
        valid = np.random.choice(idx_pool_for_valid, size=num_valid, replace=False).tolist()
    else:
        valid = []

    if test_set is None:
        test = []
        if ratio3 is None:
            num_test = N - num_train_plus_valid
        else:
            num_test = int(N * ratio3) if ratio3 <= 1.0 else int(ratio3)
    else:
        test = copy.deepcopy(test_set)
        num_test = len(test)
    #print("num_train, num_valid, num_test: ", num_train, num_valid, num_test)
    assert (num_train_plus_valid + num_test <= N)

    idx_all = list(range(N))
    if len(test) == 0:
        idx_notSelected = [i for i in idx_all if i not in idx_dedicated_for_training + valid]
        test = np.random.choice(idx_notSelected, size=num_test, replace=False).tolist()
    idx_hiPrio_for_P = [] if idx_hiPrio_for_P is None else copy.deepcopy(idx_hiPrio_for_P)
    idx_hiPrio_for_P = [_ for _ in idx_hiPrio_for_P if _ not in idx_dedicated_for_training + valid + test]
    idx_loPrio_for_P = [_ for _ in idx_all if _ not in idx_dedicated_for_training + valid + test + idx_hiPrio_for_P]

    t0 = time.time()
    dist_arr = np.ndarray([len(idx_all), len(idx_all)])
    dist_arr[:] = np.NaN
    for i in idx_all:
        dist_i_to_all = ED(X, X[i:i+1])
        dist_arr[i] = dist_i_to_all

    if len(valid) == 0:
        num_to_add = num_train_plus_valid - len(idx_dedicated_for_training)
    else:
        num_to_add = num_train - len(idx_dedicated_for_training)
    num_added = 0
    P = copy.deepcopy(idx_dedicated_for_training)
    if gamma_stratified_sampling:
        assert(X.shape[1] == 5)
        gamma_list = np.sort(np.unique(X[:, -1])).tolist()
        idx_hiPrio_for_P_list = gamma_stratified_idx(idx_hiPrio_for_P, X, gamma_list)
        idx_loPrio_for_P_list = gamma_stratified_idx(idx_loPrio_for_P, X, gamma_list)
        P_list = gamma_stratified_idx(P, X, gamma_list)
    else:
        gamma_list = [None]
        idx_hiPrio_for_P_list = [copy.deepcopy(idx_hiPrio_for_P)]
        idx_loPrio_for_P_list = [copy.deepcopy(idx_loPrio_for_P)]
        P_list = [copy.deepcopy(P)]
    while(num_added < num_to_add):
        if num_added == 0:
            #print("num_added={}, #P={}".format(num_added, len(P)))
            for i in range(len(gamma_list)):
                pass
                #print("   gamma_level={}, #hi={}, #lo={}, #_P={}".format(i, len(idx_hiPrio_for_P_list[i]), len(idx_loPrio_for_P_list[i]), len(P_list[i])))
        for _idx_hiPrio_for_P, _idx_loPrio_for_P, _P in zip(idx_hiPrio_for_P_list, idx_loPrio_for_P_list, P_list):
            if ensure_strata_balance:
                if len(_P) > np.min([len(_) for _ in P_list]):
                    continue
            pool = _idx_hiPrio_for_P if len(_idx_hiPrio_for_P) > 0 else _idx_loPrio_for_P
            assert(len(pool) > 0)

            if distance_based_P:
                if len(_P) == 0:
                    max_minDist_pool_to_P_j = np.random.choice(pool, size=None, replace=False)
                else:
                    minDists_pool_to_P = np.min(dist_arr[pool, :][:, _P], axis=1)
                    max_minDist_pool_to_P = np.max(minDists_pool_to_P)
                    candidates_with_same_dist = np.array(pool)[
                        np.where(minDists_pool_to_P == max_minDist_pool_to_P)].tolist()
                    assert (len(candidates_with_same_dist) > 0)
                    max_minDist_pool_to_P_j = np.random.choice(candidates_with_same_dist, size=None, replace=False)
            else:
                max_minDist_pool_to_P_j = np.random.choice(pool, size=None, replace=False)
            _P.append(max_minDist_pool_to_P_j)
            pool.remove(max_minDist_pool_to_P_j)
            P.append(max_minDist_pool_to_P_j)
            num_added += 1
            if num_added == num_to_add:
                break
        if num_added == num_to_add:
            #print("num_added={}, #P={}".format(num_added, len(P)))
            for i in range(len(gamma_list)):
                pass
                #print("   gamma_level={}, #hi={}, #lo={}, #_P={}".format(i, len(idx_hiPrio_for_P_list[i]), len(idx_loPrio_for_P_list[i]), len(P_list[i])))

    assert (len(P) == num_train_plus_valid if len(valid)==0 else num_train)
    #print("np.mean(P), np.mean(X[P]):", np.mean(P), np.mean(X[P]))
    #print("np.mean(P[:300]), np.mean(X[P[:300]]):", np.mean(P[:300]), np.mean(X[P[:300]]))
    #print("P[-10:]:", P[-10:])
    #print("runtime for deterministic splitting:", time.time() - t0, "sec.")

    if len(valid) == 0:
        nonDedicated_in_P = [i for i in P if i not in idx_dedicated_for_training]
        if deterministicSplitMode == 1:
            if seed_trainValidSplit is not None:
                np.random.seed(seed_trainValidSplit)
                #print(f"Set seed={seed_trainValidSplit} for train/valid splitting.")
            valid = np.random.choice(nonDedicated_in_P, size=num_valid, replace=False).tolist()
        elif deterministicSplitMode == 2:
            print("last {} in P as valid.".format(num_valid))
            valid = nonDedicated_in_P[-num_valid:]
        elif deterministicSplitMode == 3:
            print("first {} in P as valid.".format(num_valid))
            valid = nonDedicated_in_P[:num_valid]
        else:
            valid = None
        assert(valid is not None)

        train = [i for i in P if i not in valid]
    else:
        train = P
    #print("np.mean(train):", np.mean(train))
    #print("np.mean(valid):", np.mean(valid))

    idx_train, idx_valid, idx_test = np.array(train), np.array(valid), np.array(test)

    #print(np.shape(idx_train), np.shape(idx_valid), np.shape(idx_test))
    #print(np.mean(idx_train), np.mean(idx_valid), np.mean(idx_test))
    X_train, X_valid, X_test = X[idx_train], X[idx_valid], X[idx_test]
    y_train, y_valid, y_test = y[idx_train], y[idx_valid], y[idx_test]
    #print(X_train.mean(), X_valid.mean(), X_test.mean())
    return X_train, X_valid, X_test, y_train, y_valid, y_test, idx_train, idx_valid, idx_test


# remember the seed, then we'r able to re-do the same random splitting using the same seed
def train_test_split_v3(X, y,
                        ratio, ratio2, ratio3=None,
                        idx_to_skip=None,
                        idx_dedicated_for_training=None,
                        keep_train_amount=False,
                        use_all_for_training=False,
                        test_set=None,
                        stratified_sampling=False,
                        stratified_base=None,
                        seed=None):
    assert(X.shape[0] == y.shape[0])
    N, D = X.shape
    if stratified_base is None:
        stratified_base = X
    else:
        assert(stratified_base.shape[0] == X.shape[0])
    if idx_dedicated_for_training is None:
        idx_dedicated_for_training = []

    num_train = int(N * ratio) if ratio <= 1.0 else int(ratio)
    num_valid = int(N * ratio2) if ratio2 <= 1.0 else int(ratio2)
    if test_set is None:
        test = []
        if ratio3 is None:
            num_test = N - num_train - num_valid
        else:
            num_test = int(N * ratio3) if ratio3 <= 1.0 else int(ratio3)
    else:
        test = copy.deepcopy(test_set)
        num_test = len(test)
    print("num_train, num_valid, num_test: ", num_train, num_valid, num_test)
    assert (num_train + num_valid + num_test <= N)
    assert (num_train >= len(idx_dedicated_for_training))

    if seed is not None:
        np.random.seed(seed)
    idx_all = list(range(N))
    idx_all_except_dedicated_shuffled = [i for i in idx_all if i not in idx_dedicated_for_training + test]
    np.random.shuffle(idx_all_except_dedicated_shuffled)
    N2 = len(idx_all_except_dedicated_shuffled)

    if stratified_sampling:
        D2 = stratified_base.shape[1]
        if D2 > 5:
            print("!!!!! [WARN] !!!!!: strata dim > 5, change to 5 from old value", D)
            D2 = 5
        # build up strata dict
        xmin, xmax = np.min(stratified_base, axis=0), np.max(stratified_base, axis=0)
        xmiddle = (xmin + xmax) / 2.0
        print("xmin:", xmin)
        print("xmax:", xmax)
        print("xmiddle:", xmiddle)
        strata_dict = {}
        stratum_names = []  # use list rather than set, in order to keep consistent through different runs
        for i in range(N2):
            idx_tmp = idx_all_except_dedicated_shuffled[i]
            stratum = ""
            for j in range(D2):
                if stratified_base[idx_tmp, j] < xmiddle[j]:
                    stratum += "s1"
                else:
                    stratum += "s2"
            if stratum not in strata_dict:
                strata_dict[stratum] = [idx_tmp]
            else:
                strata_dict[stratum].append(idx_tmp)
            if stratum not in stratum_names:
                stratum_names.append(stratum)
        stratum_N = len(stratum_names)
        print("stratum_N:", stratum_N)
        vs = [len(v) for v in strata_dict.values()]
        print("sum/max/min/mean/median # in strata:", np.sum(vs), np.max(vs), np.min(vs), np.mean(vs), np.median(vs))
        for k,v in strata_dict.items():
            print("strata and percentage:", k, len(v), len(v) / N2)

        # sampling
        np.random.shuffle(stratum_names)
        for stratum in stratum_names:
            np.random.shuffle(strata_dict[stratum])

        if len(test) == 0:
            idx_test = []
            strata_cnt_test = {k:0 for k in stratum_names}
            i = np.random.choice(len(stratum_names), replace=False)
            while(len(idx_test) < num_test):
                stratum_selected = stratum_names[i]
                if len(strata_dict[stratum_selected]) >= 1:
                    idx_test.append(strata_dict[stratum_selected].pop())
                    strata_cnt_test[stratum_selected] += 1
                i += 1
                if i >= len(stratum_names):
                    i = 0
            print("(test set) #strata:", len(strata_cnt_test.keys()))
            vs = list(strata_cnt_test.values())
            print("(test set) sum/max/min # in strata:", np.sum(vs), np.max(vs), np.min(vs))
            if num_test % stratum_N == 0:
                assert(np.max(vs) - np.min(vs) == 0)
            else:
                assert (np.max(vs) - np.min(vs) <= 1)
        else:
            idx_test = test

        idx_valid = []
        strata_cnt_valid = {k:0 for k in stratum_names}
        i = np.random.choice(len(stratum_names), replace=False)
        while(len(idx_valid) < num_valid):
            stratum_selected = stratum_names[i]
            if len(strata_dict[stratum_selected]) >= 1:
                idx_valid.append(strata_dict[stratum_selected].pop())
                strata_cnt_valid[stratum_selected] += 1
            i += 1
            if i >= len(stratum_names):
                i = 0
        print("(valid set) #strata:", len(strata_cnt_valid.keys()))
        vs = list(strata_cnt_valid.values())
        print("(valid set) sum/max/min # in strata:", np.sum(vs), np.max(vs), np.min(vs))
        '''
        if num_valid % stratum_N == 0:
            assert (np.max(vs) - np.min(vs) == 0)
        else:
            assert (np.max(vs) - np.min(vs) <= 1)
        '''
    else:
        if len(test) == 0:
            idx_test  = idx_all_except_dedicated_shuffled[0:num_test]
            idx_valid = idx_all_except_dedicated_shuffled[num_test : num_test+num_valid]
        else:
            idx_test = test
            idx_valid = idx_all_except_dedicated_shuffled[0:num_valid]

    idx_valid_test = idx_valid + idx_test
    idx_train_pool = [i for i in idx_all if i not in idx_valid_test]
    np.random.shuffle(idx_train_pool) # must shuffle otherwise only the first samples are used for training.
    print("train: #dedicated & #randomly:", len(idx_dedicated_for_training), num_train - len(idx_dedicated_for_training))
    idx_extra = idx_train_pool[0 : num_train - len(idx_dedicated_for_training)]
    idx_train = idx_dedicated_for_training + idx_extra
    np.random.shuffle(idx_train)

    idx_train, idx_valid, idx_test = np.array(idx_train), np.array(idx_valid), np.array(idx_test)

    #print(np.shape(idx_train), np.shape(idx_valid), np.shape(idx_test))
    X_train, X_valid, X_test = X[idx_train], X[idx_valid], X[idx_test]
    y_train, y_valid, y_test = y[idx_train], y[idx_valid], y[idx_test]
    return X_train, X_valid, X_test, y_train, y_valid, y_test, idx_train, idx_valid, idx_test

# remember the seed, then we'r able to re-do the same random splitting using the same seed
def train_test_split_v2(X, y, ratio, ratio2, ratio3=None, idx_to_skip=None, idx_dedicated_for_training=None, keep_train_amount=False, use_all_for_training=False, seed=None):
    assert(X.shape[0] == y.shape[0])
    if seed is not None:
        np.random.seed(seed)
    #print("np.random.get_state():\n", np.random.get_state())
    idx = np.random.permutation(X.shape[0])
    num_train = int(X.shape[0] * ratio)
    num_valid = int(X.shape[0] * ratio2)
    if ratio3 is None:
        idx_train, idx_valid, idx_test = idx[0:num_train], idx[num_train:num_train+num_valid], idx[num_train+num_valid:]
    else:
        assert(ratio + ratio2 + ratio3 <= 1.0)
        num_test = int(X.shape[0] * ratio3)
        idx_train, idx_valid, idx_test = idx[0:num_train], idx[num_train:num_train+num_valid], idx[num_train+num_valid:num_train+num_valid+num_test]
    if idx_to_skip is not None:
        idx_train = skipSomeIdx(idx_train, idx_to_skip)
        idx_valid = skipSomeIdx(idx_valid, idx_to_skip)
        #skipSomeIdx(idx_test, idx_to_skip)
    if use_all_for_training:
        print("Before adding valid & test into training: idx_train.shape=", idx_train.shape)
        idx_train = np.concatenate([idx_train, idx_valid, idx_test])
        print("After adding valid & test into training: idx_train.shape=", idx_train.shape)
    if idx_dedicated_for_training is not None:
        print("Before adding dedicated idx: idx_train.shape={}, idx_valid.shape={}, idx_test.shape={}".format(idx_train.shape, idx_valid.shape, idx_test.shape))
        idx_train, train_added_num   = addSomeIdx(idx_train, idx_dedicated_for_training)
        idx_valid, valid_removed_num = removeSomeIdx(idx_valid, idx_dedicated_for_training)
        idx_test,  test_removed_num  = removeSomeIdx(idx_test,  idx_dedicated_for_training)
        print("After adding dedicated idx: idx_train.shape={}, idx_valid.shape={}, idx_test.shape={}".format(idx_train.shape, idx_valid.shape, idx_test.shape))
        if keep_train_amount:
            idx_train, idx_train_removed = remove_n_idx(idx_train, train_added_num, except_idx=idx_dedicated_for_training)
            idx_pool = idx_train_removed
            idx_valid = np.concatenate([idx_valid, idx_pool[:valid_removed_num]])
            idx_test = np.concatenate([idx_test, idx_pool[valid_removed_num : valid_removed_num + test_removed_num]])
            print("After keeping orig amounts: idx_train.shape={}, idx_valid.shape={}, idx_test.shape={}".format(idx_train.shape, idx_valid.shape, idx_test.shape))
    print(np.shape(idx_train), np.shape(idx_valid), np.shape(idx_test))
    X_train, X_valid, X_test = X[idx_train], X[idx_valid], X[idx_test]
    y_train, y_valid, y_test = y[idx_train], y[idx_valid], y[idx_test]
    return X_train, X_valid, X_test, y_train, y_valid, y_test, idx_train, idx_valid, idx_test

# remember the seed, then we'r able to re-do the same random splitting using the same seed
def data_split(cmykt_arr, spectral_transl_arr, lab_arr, ratio, ratio2=None, returnIdx=False, seed=123):
    X_train, X_test, y_train, y_test, idx_train, idx_test = train_test_split(cmykt_arr, spectral_transl_arr, ratio, ratio2=ratio2, seed=seed)
    cmykt_train, cmykt_test = X_train, X_test
    spec_train, spec_test = y_train, y_test
    lab_train, lab_test = lab_arr[idx_train], lab_arr[idx_test] # must use the same indexes otherwise lab will mismatch with cmykt!
    if returnIdx:
        return cmykt_train, cmykt_test, spec_train, spec_test, lab_train, lab_test, idx_train, idx_test
    else:
        return cmykt_train, cmykt_test, spec_train, spec_test, lab_train, lab_test
# remember the seed, then we'r able to re-do the same random splitting using the same seed

def data_split_v3(X, y, ratio, ratio2, ratio3=None, idx_to_skip=None, idx_dedicated_for_training=None,
                  keep_train_amount=False, use_all_for_training=False, returnIdx=False, stratified_sampling=True, stratified_base=None, seed=None, seed_trainValidSplit=None, idx_hiPrio_for_P=None, distance_based_P=True, idx_pool_for_valid=None, gamma_stratified_sampling=False, ensure_strata_balance=False, deterministicSplit=False, deterministicSplitMode=1, test_set=None):
    if not deterministicSplit:
        X_train, X_valid, X_test, y_train, y_valid, y_test, idx_train, idx_valid, idx_test = train_test_split_v3(X, y,
                                                                                                             ratio,
                                                                                                             ratio2,
                                                                                                             ratio3=ratio3,
                                                                                                             idx_to_skip=idx_to_skip,
                                                                                                             idx_dedicated_for_training=idx_dedicated_for_training,
                                                                                                             keep_train_amount=keep_train_amount,
                                                                                                             test_set=test_set,
                                                                                                             use_all_for_training=use_all_for_training,
                                                                                                             stratified_sampling=stratified_sampling,
                                                                                                             stratified_base=stratified_base,
                                                                                                             seed=seed)
    else:
        #print("Deterministic data splitting.")
        X_train, X_valid, X_test, y_train, y_valid, y_test, idx_train, idx_valid, idx_test = train_test_split_v4(X, y,
                                                                                                             ratio,
                                                                                                             ratio2,
                                                                                                             ratio3=ratio3,
                                                                                                             idx_dedicated_for_training=idx_dedicated_for_training,
                                                                                                             test_set=test_set,
                                                                                                             idx_hiPrio_for_P=idx_hiPrio_for_P,
                                                                                                             distance_based_P=distance_based_P,
                                                                                                             idx_pool_for_valid=idx_pool_for_valid,
                                                                                                             gamma_stratified_sampling=gamma_stratified_sampling,
                                                                                                             ensure_strata_balance=ensure_strata_balance,
                                                                                                             deterministicSplitMode=deterministicSplitMode,
                                                                                                             seed=seed,
                                                                                                             seed_trainValidSplit=seed_trainValidSplit)
    if returnIdx:
        return X_train, X_valid, X_test, y_train, y_valid, y_test, idx_train, idx_valid, idx_test
    else:
        return X_train, X_valid, X_test, y_train, y_valid, y_test


def data_split_v2(X, y, ratio, ratio2, ratio3=None, idx_to_skip=None, idx_dedicated_for_training=None, keep_train_amount=False, use_all_for_training=False, returnIdx=False, seed=None):
    X_train, X_valid, X_test, y_train, y_valid, y_test, idx_train, idx_valid, idx_test = train_test_split_v2(X, y, ratio, ratio2, ratio3=ratio3,
                                                                                                             idx_to_skip=idx_to_skip,
                                                                                                             idx_dedicated_for_training=idx_dedicated_for_training,
                                                                                                             keep_train_amount=keep_train_amount,
                                                                                                             use_all_for_training = use_all_for_training,
                                                                                                             seed=seed)
    if returnIdx:
        return X_train, X_valid, X_test, y_train, y_valid, y_test, idx_train, idx_valid, idx_test
    else:
        return X_train, X_valid, X_test, y_train, y_valid, y_test


def save_data_split_indexes(idx_filepath, idx_train, idx_valid, idx_test):
    l1, l2, l3 = idx_train.shape[0], idx_valid.shape[0], idx_test.shape[0]
    idx_all = np.concatenate([np.array([l1, l2, l3]), idx_train, idx_valid, idx_test])
    np.savetxt(idx_filepath, idx_all, fmt='%d')

def load_data_split_indexes(idx_filepath):
    txt = np.loadtxt(idx_filepath, dtype=int)
    l1, l2, l3 = txt[0], txt[1], txt[2]
    idx_train, idx_valid, idx_test = txt[3:3+l1], txt[3+l1:3+l1+l2], txt[3+l1+l2:3+l1+l2+l3]
    return idx_train, idx_valid, idx_test


def scaleDownLab(lab):
    '''
    lab = copy.deepcopy(lab0)
    lab[:, 0] /= 100.
    lab[:, 1] = lab[:, 1] / (128 * 2) + 0.5
    lab[:, 2] = lab[:, 2] / (128 * 2) + 0.5
    '''
    a1 = np.ones_like(lab[:,0:1]) * 0.
    a2 = np.ones_like(lab[:,0:1]) * 0.5
    a3 = np.ones_like(lab[:,0:1]) * 0.5
    a = np.concatenate([a1,a2,a3], 1)
    print(a.shape)
    b1 = np.ones_like(lab[:,0:1]) * 100.
    b2 = np.ones_like(lab[:,0:1]) * (128*2)
    b3 = np.ones_like(lab[:,0:1]) * (128*2)
    b = np.concatenate([b1,b2,b3], 1)
    return lab / b + a
def scaleUpLab(lab):
    '''
    lab = copy.deepcopy(lab0)
    lab[:, 0] *= 100
    lab[:, 1] = (lab[:, 1] - 0.5) * (128 * 2)
    lab[:, 2] = (lab[:, 2] - 0.5) * (128 * 2)
    '''
    a1 = np.ones_like(lab[:, 0:1]) * 0.
    a2 = np.ones_like(lab[:, 0:1]) * 0.5
    a3 = np.ones_like(lab[:, 0:1]) * 0.5
    a = np.concatenate([a1, a2, a3], 1)
    print(a.shape)
    b1 = np.ones_like(lab[:, 0:1]) * 100.
    b2 = np.ones_like(lab[:, 0:1]) * (128 * 2)
    b3 = np.ones_like(lab[:, 0:1]) * (128 * 2)
    b = np.concatenate([b1, b2, b3], 1)
    return (lab - a) * b


def sigmoid(x):
    return 1 / (1 + np.power(np.e, -x))
def inv_sigmoid(y):
    assert((y<1.0).all() and (y>0.0).all()), print(f"y range [{np.min(y)}, {np.max(y)}] is out of range [0.,1.]")
    return np.log(y/(1-y))
def shifted_sigmoid(x):
    return 1 / (1 + np.power(np.e, -(x + inv_sigmoid(0.135)))) # shift to the right so that y(x=0)=0.135
def sigmoid_derivative(x):
    sig = sigmoid(x)
    return sig * (1 - sig)


def selectFromList(a_list, idx_selected):
    a_list_2 = []
    for ii in idx_selected:
        a_list_2.append(a_list[ii])
    return a_list_2


def roundAsList(a, d=3):
    return list(np.round(a, d))

def roundAsArray(a, d=3):
    return np.round(a, d)

def write_formally(fname, data, d=6):
    assert(len(data.shape) == 2)
    row, col = data.shape
    print("decimal:", d)
    fmt = "{" + ":.{}f".format(d) + "}"
    with open(fname, 'w') as f:
        f.write("ROWS\t{}\n".format(row))
        f.write("COLUMNS\t{}\n".format(col))
        f.write("BEGIN_DATA\n")
        for i in range(row):
            line = '\t'.join([fmt.format(tmp) for tmp in data[i]])
            f.write("{}\n".format(line))
        f.write("END_DATA")
    print("Saved as", fname)

