from util.cost_util import * # MSE, RMS, plot_E_lab, plot_abs_distri
from util.data_processing import *

import tensorflow as tf
'''
import tensorflow
if tensorflow.__version__.startswith("1."):
    import tensorflow as tf
else:
    import tensorflow.compat.v1 as tf
    tf.disable_v2_behavior()
'''
#print('tf.executing_eagerly():', tf.executing_eagerly())

from tensorflow import keras
from tensorflow.keras import layers, regularizers
from tensorflow.keras import backend as K
from tensorflow.keras.utils import plot_model
import gc

import time
from datetime import datetime

class ModelTrainConfig:
    ns_dict = {
        'DLLCN': [200, 400, 800, 400, 200],
        'PDL': [200, 1500, 1500, 1500, 200]
    }
    dropout_dict = {
        'DLLCN': 0.1,
        'PDL': 0.05
    }
    output_descr_dict = {
        'stratasys': {
            'DLLCN': 'cmykg',
            'PDL': 'specAndAlpha'
        },
        'mimaki_1': {
            'PDL': 'specAndAlpha'
        },
        'mimaki_2': {
            'DLLCN': 'cmykg',
            'PDL': 'spec'
        },
    }
    activation_final_dict = {
        'DLLCN': 'linear',
        'PDL': 'sigmoid_hshift'
    }
    def __init__(self, dataset, data, approach):
        self.dataset = dataset
        self.data = data
        self.approach = approach

        ############## model settings ###############
        self.output_descr = self.output_descr_dict[self.dataset][self.approach]
        self.ns = self.ns_dict[self.approach]
        if self.approach == 'DLLCN':
            self.hlayer_split = None
            self.ns2 = None
            self.n_out2 = None
            self.n_out = 4 if self.dataset == 'mimaki_2' else 5
        elif self.approach == 'PDL':
            if self.dataset == 'mimaki_2':
                self.hlayer_split = None
                self.ns2 = None
                self.n_out2 = None
                self.n_out = None
            else:
                self.hlayer_split = -2
                self.ns2 = [30]
                self.n_out2 = 1
                self.n_out = 31
        self.cmykgPredicted = True if self.approach == 'DLLCN' else False
        self.CMYKG2cmykg = True if self.approach == 'DLLCN' else False
        self.cmykg_clipAfterMul = True if self.approach == 'DLLCN' else False
        self.cmykg_sigmoidAfterMul = False

        self.dropout = self.dropout_dict[self.approach]

        self.weight_initializer = "glorot_normal"
        self.lrelu_alpha = 0.01
        self.activation_final = self.activation_final_dict[self.approach]
        self.loss_obj = AppearanceLossPool(labPredicted=False, specPredicted=True, alphaPredicted=True, cmykgPredicted=self.cmykgPredicted,
                                          lab_loss=ColorDiff.ciede2000noMean_tensor, spec_loss=PrimitiveLossPool.RMSE_tensor, alpha_loss=PrimitiveLossPool.MAE_tensor,
                                          Elab_weight=1.0, Espec_weight=20.0, Ealpha_weight=10.0,
                                          interp_ref=self.data.get_CN_ref_values(),
                                          cmykg_sigmoidAfterMul=self.cmykg_sigmoidAfterMul,
                                          cmykg_clipAfterMul=self.cmykg_clipAfterMul,
                                          dbg=False)
        self.loss_fn = self.loss_obj.get_loss_func()

        ############## training settings ###############
        self.bs = None
        self.lr = 0.009
        self.ep = 15000
        self.callbacks = [
            keras.callbacks.ReduceLROnPlateau(monitor='val_loss', factor=1. / 3., patience=1000, min_lr=0, verbose=0),
            MyEarlyStopping(patience=2000, train_loss_weight=0.0, filepath='./model_tmp/model')
        ]

        ############## data settings ###############

class scaler_layer(object):
    def __init__(self, scales):
        self.scales = scales

    def scaler_fn(self, x):
        for i, scale in enumerate(self.scales):
            x[:, i] = x[:, i] * (scale[1] - scale[0]) + scale[0]
        return x

    def scaler_output_shape_fn(self, input_shape):
        return input_shape


def scaler_fn(x, scales):
    for i, scale in enumerate(range(len(scales))):
        x[:, i] = x[:, i] * (scale[1] - scale[0]) + scale[0]
    return x


def scaler_layer_output_shape(input_shape):
    return input_shape


class SigmoidHshift(tf.keras.layers.Layer):

    def __init__(self, hshift_amount, name="sigmoid_hshift", **kwargs):
        super(SigmoidHshift, self).__init__(name=name, **kwargs)
        self.hshift_amount = hshift_amount
        #print("__init__: self.hshift_amount:", self.hshift_amount)

    def call(self, inputs):
        #print("call: inputs.shape:", inputs.shape)
        #print("call: self.hshift_amount:", self.hshift_amount)
        return tf.keras.activations.sigmoid(inputs + self.hshift_amount)

    def get_config(self):
        config = super(SigmoidHshift, self).get_config()
        config.update({'hshift_amount': self.hshift_amount})
        #print("get_config: self.hshift_amount:", self.hshift_amount)
        return config

    @classmethod
    def from_config(cls, config):
        return cls(**config)


def get_activation(act_fn_str, lrelu_alpha=0.01, sigmoid_hscale_factor=1.0, sigmoid_hshift_amount=0.0):
    def leaky_relu(x):
        #print('lrelu_alpha:', lrelu_alpha)
        return keras.activations.relu(x, alpha=lrelu_alpha, max_value=None, threshold=0.0)

    def sigmoid_hscale(x):
        print('hscale factor:', sigmoid_hscale_factor)
        return keras.activations.sigmoid(x / sigmoid_hscale_factor)

    def sigmoid_hshift(x):
        return tf.keras.activations.sigmoid(x + sigmoid_hshift_amount)

    if act_fn_str == 'relu':
        return leaky_relu
    elif act_fn_str == 'sigmoid_hscale':
        return sigmoid_hscale
    elif act_fn_str == 'sigmoid_hshift':
        return sigmoid_hshift
    else:
        return act_fn_str


def build_model(n_in, ns, n_out,
                base_net=None, base_net_direct_use=False,
                CMYKG2cmykg=False, cmykg_noMul=False, cmykg_addOne=False,
                hlayer_split=-2, ns2=None, n_out2=None,
                activation='relu', activation_final='linear',
                lrelu_alpha=0.01,
                sigmoid_hscale_factor = 1.0,
                sigmoid_hshift_amount = 0.0,
                kernel_initializer="glorot_normal",
                kernel_initializer_final="glorot_normal",
                kernel_init_seed=None,
                scales=None,
                BN=False, BN_on_input=False,
                dropout=0., dropout_on_input=0., reg_l1=0., reg_l2=0.,
                clipnorm=None,
                loss='mse',
                run_eagerly=False):
    if base_net_direct_use:
        model = tf.keras.models.clone_model(base_net)
        model.set_weights(base_net.get_weights())
        if clipnorm is not None:
            print("clipnorm=", clipnorm)
            optimizer = keras.optimizers.Adam(clipnorm=clipnorm)
        else:
            optimizer = keras.optimizers.Adam()
        model.compile(optimizer=optimizer, loss=loss, run_eagerly=run_eagerly)
        #print("model compiled with run_eagerly =", run_eagerly)
        return model

    reg = regularizers.l1_l2(l1=reg_l1, l2=reg_l2)

    if kernel_initializer == "glorot_normal":
        kernel_initializer = tf.keras.initializers.glorot_normal(seed=kernel_init_seed)  # tf.initializers.GlorotNormal(seed=kernel_init_seed)
    if kernel_initializer_final == "glorot_normal":
        kernel_initializer_final = tf.keras.initializers.glorot_normal(seed=kernel_init_seed)  # tf.initializers.GlorotNormal(seed=kernel_init_seed)
    #print("kernel_initializer:", kernel_initializer)
    #print("kernel_initializer_final:", kernel_initializer_final)

    # input layer
    inputs = keras.Input(shape=(n_in,))
    if base_net is not None:
        out = base_net(inputs)
        print("Added base_net:")
        print(base_net.summary())
    else:
        out = inputs
    if BN_on_input:
        out = tf.keras.layers.BatchNormalization(axis=-1, momentum=0.99, epsilon=0.001, center=True, scale=True)(out)
    if dropout_on_input:
        out = tf.keras.layers.Dropout(dropout_on_input)(out)

    # hidden layers -- path 1
    hidden_layers = []
    for n in ns:
        out = layers.Dense(n, activation=activation, kernel_initializer=kernel_initializer, kernel_regularizer=reg)(out)
        if BN:
            out = tf.keras.layers.BatchNormalization(axis=-1, momentum=0.99, epsilon=0.001, center=True, scale=True)(out)
        if dropout:
            out = tf.keras.layers.Dropout(dropout)(out)
        hidden_layers.append(out)
    # hidden layers -- path 2
    out2 = None
    if ns2 is not None:
        out2 = hidden_layers[hlayer_split]
        for n in ns2:
            out2 = layers.Dense(n, activation=activation, kernel_initializer=kernel_initializer, kernel_regularizer=reg)(out2)
            if BN:
                out2 = tf.keras.layers.BatchNormalization(axis=-1, momentum=0.99, epsilon=0.001, center=True, scale=True)(out2)
            if dropout:
                out2 = tf.keras.layers.Dropout(dropout)(out2)


    # final layer
    # path 1
    out = tf.keras.layers.Dense(n_out, activation=None, kernel_initializer=kernel_initializer_final, kernel_regularizer=reg)(out)
    # path 2
    if out2 is not None:
        out2 = tf.keras.layers.Dense(n_out2, activation=None, kernel_initializer=kernel_initializer_final, kernel_regularizer=reg)(out2)
    # merge 2 paths
    if out2 is not None:
        #print("Before merge, out.shape:", out.shape)
        out = tf.keras.layers.concatenate([out, out2], axis=-1)
        #print("After merge, out.shape:", out.shape)
    if activation_final == "sigmoid_hshift":
        #print("use SigmoidHshift.")
        #print('hshift amount:', sigmoid_hshift_amount)
        out = SigmoidHshift(sigmoid_hshift_amount)(out)
    elif activation_final == "relu" or activation_final == "lrelu":
        #print("use LeakyReLU.")
        out = tf.keras.layers.LeakyReLU(alpha=lrelu_alpha)(out)
    else:
        #print("use regular act:", activation_final)
        out = tf.keras.layers.Activation(activation_final, name=activation_final)(out)

    # scale layer
    if scales is not None:
        scaler = scaler_layer(scales)
        out = tf.keras.layers.Lambda(scaler.scaler_fn, scaler.scaler_output_shape_fn)(out)

    # CMYKG -> cmykg
    if CMYKG2cmykg and not cmykg_noMul:
        if cmykg_addOne:
            out += 1.0
        #out = [out, inputs]
        out = tf.keras.layers.concatenate([out, inputs], axis=-1)
        #print("out.shape:", out.shape)

    # instantiate the model
    model = tf.keras.Model(inputs=inputs, outputs=out)

    # compile
    if clipnorm is not None:
        print("clipnorm=", clipnorm)
        optimizer = keras.optimizers.Adam(clipnorm=clipnorm)
    else:
        optimizer = keras.optimizers.Adam()
    model.compile(optimizer=optimizer, loss=loss, run_eagerly=run_eagerly)
    #print("model compiled with run_eagerly =", run_eagerly)
    return model


from colormath.color_objects import LabColor, XYZColor
from colormath.color_diff import delta_e_cie1976, delta_e_cie1994, delta_e_cie2000
def color_diff(lab1, lab2, type='cie2000', Kl=1, Kc=1, Kh=1):
    assert(lab1.shape == lab2.shape)
    assert(len(lab1.shape) == 1 or len(lab1.shape) == 2)
    if len(lab1.shape) == 1:
        color1 = LabColor(lab_l=lab1[0], lab_a=lab1[1], lab_b=lab1[2])
        color2 = LabColor(lab_l=lab2[0], lab_a=lab2[1], lab_b=lab2[2])
        if type == 'cie2000':
            color_diff = delta_e_cie2000(color1, color2, Kl=Kl, Kc=Kc, Kh=Kh)
        elif type == 'cie1994':
            color_diff = delta_e_cie1994(color1, color2, K_L=Kl, K_C=Kc, K_H=Kh)
        elif type == 'cie1976':
            color_diff = delta_e_cie1976(color1, color2)
        else:
            sys.exit('unknown color type!')
        return color_diff
    else:
        color_diffs = np.ndarray(shape=[lab1.shape[0],1])
        for i in range(lab1.shape[0]):
            color1 = LabColor(lab_l=lab1[i, 0], lab_a=lab1[i, 1], lab_b=lab1[i, 2])
            color2 = LabColor(lab_l=lab2[i, 0], lab_a=lab2[i, 1], lab_b=lab2[i, 2])
            if type == 'cie2000':
                color_diffs[i] = delta_e_cie2000(color1, color2, Kl=Kl, Kc=Kc, Kh=Kh)
            elif type == 'cie1994':
                color_diffs[i] = delta_e_cie1994(color1, color2, K_L=Kl, K_C=Kc, K_H=Kh)
            elif type == 'cie1976':
                color_diffs[i] = delta_e_cie1976(color1, color2)
            else:
                sys.exit('unknown color type!')
        return color_diffs


import numpy


def my_delta_e_cie1994(lab1, lab2,
                       K_L=1, K_C=1, K_H=1, K_1=0.045, K_2=0.015):
    """
    Calculates the Delta E (CIE1994) of two colors.
    K_l:
      0.045 graphic arts
      0.048 textiles
    K_2:
      0.015 graphic arts
      0.014 textiles
    K_L:
      1 default
      2 textiles
    """
    C_1 = numpy.sqrt(numpy.sum(numpy.power(lab1[:, 1:], 2), axis=1))
    C_2 = numpy.sqrt(numpy.sum(numpy.power(lab2[:, 1:], 2), axis=1))

    delta_lab = lab1 - lab2

    delta_L = delta_lab[:, 0].copy()
    delta_C = C_1 - C_2
    delta_lab[:, 0] = delta_C

    delta_H_sq = numpy.sum(numpy.power(delta_lab, 2) * numpy.array([-1, 1, 1]), axis=1)
    # noinspection PyArgumentList
    delta_H = numpy.sqrt(delta_H_sq.clip(min=0))

    S_L = 1
    S_C = 1 + K_1 * C_1
    S_H = 1 + K_2 * C_1

    LCH = numpy.vstack([delta_L, delta_C, delta_H])
    params = numpy.array([[K_L * S_L], [K_C * S_C], [K_H * S_H]])
    print(LCH.shape, params.shape)
    print(LCH)
    print(params)

    print((numpy.power(LCH / params, 2)))

    return numpy.sqrt(numpy.sum(numpy.power(LCH / params, 2), axis=0))


'''
K.clear_session()
gc.collect()

# feed dict
X_train, y_train = cmykt_train_poly, lab_train
X_test, y_test = cmykt_test_poly, lab_test
print(X_train.shape, y_train.shape)

# build a NN model
n_in = X_train.shape[1]
n_out = y_train.shape[1]
ns = [1500, 1500, 1500]
activation = 'relu'
activation_final = 'linear'
NN_tmp = build_model(n_in, ns, n_out, activation, activation_final)
NN_tmp.summary()

# train model
lr = 1 * 1e-3
K.set_value(NN_tmp.optimizer.lr, lr)
bs = y_train.shape[0]
ep = 3000
callbackslist = [
    keras.callbacks.EarlyStopping(monitor='val_loss', patience=1000,
                                  min_delta=1e-5, baseline=None, restore_best_weights=True),
    keras.callbacks.ModelCheckpoint(filepath='NN.h5', monitor='val_loss', save_best_only=True,
                                    save_weights_only=True, period=1),
    #tf.keras.callbacks.TensorBoard(log_dir=os.path.join('log_dir'), histogram_freq=0, batch_size=32,
    #                            write_graph=True, write_grads=False, write_images=False, update_freq='epoch')
]
t0 = time.time()
history = NN_tmp.fit(X_train, y_train, epochs=ep, batch_size=bs,
                     validation_data=(X_test, y_test),
                    callbacks=callbackslist)
print("training runtime: ", time.time() - t0)

# plot training history
print(ns, lr, np.min(history.history['loss']), np.min(history.history['val_loss']))
plt.plot(history.history['loss'])
plt.plot(history.history['val_loss'])
plt.title('MSE loss over epoches')
plt.legend(['train', 'val'])
plt.yscale('log')
plt.show()
'''


def build_and_train(X_train, y_train, X_test, y_test,
                    clear_session=True,
                    ns=[1500, 1500, 1500],
                    activation='relu', activation_final='linear',
                    kernel_initializer="glorot_normal",
                    scales=None,
                    BN=False,
                    dropout=0., reg_l1=0., reg_l2=0., loss='mse',
                    pretrained_model=None, load_weights=False, checkpoint_to_load='NN.h5',
                    lr=1e-3, bs=8, ep=3000, train_verbose=1,
                    clipnorm=None,
                    earlyStop=True, earlyStop_patience=1000, earlyStop_verbose=0,
                    reduceLR=True, reduceLR_factor=0.3, reduceLR_patience=1000, reduceLR_verbose=0,
                    saveModel=True, save_path='NN.h5', save_weights_only=True, save_best_only=True,
                    tensorBoard=False, log_dir='log_dir', update_freq='epoch',
                    restore_best=True,
                    plot_history=True,
                    show_Elab=False,
                    show_ED=False
                    ):
    if clear_session:
        K.clear_session()
        gc.collect()

    # build a NN model
    if pretrained_model is None:
        n_in = X_train.shape[1]
        n_out = y_train.shape[1]
        NN_tmp = build_model(n_in, ns, n_out, activation, activation_final,
                             kernel_initializer=kernel_initializer,
                             scales=scales, BN=BN,
                             clipnorm=clipnorm,
                             dropout=dropout, reg_l1=reg_l1, reg_l2=reg_l2, loss=loss)
    else:
        NN_tmp = pretrained_model

    if load_weights:
        NN_tmp.load_weights(checkpoint_to_load)

    # train model
    K.set_value(NN_tmp.optimizer.lr, lr)
    callbackslist = []
    if earlyStop:
        callbackslist.append(
            keras.callbacks.EarlyStopping(monitor='val_loss', patience=earlyStop_patience,
                                          min_delta=1e-5, baseline=None, verbose=earlyStop_verbose)
            # , restore_best_weights=False)
        )
    if reduceLR:
        callbackslist.append(
            keras.callbacks.ReduceLROnPlateau(monitor='val_loss', factor=reduceLR_factor, patience=reduceLR_patience,
                                              verbose=reduceLR_verbose)
        )
    if saveModel:
        callbackslist.append(
            keras.callbacks.ModelCheckpoint(filepath=save_path, monitor='val_loss', save_best_only=save_best_only,
                                            save_weights_only=save_weights_only, period=1),
        )
    if tensorBoard:
        callbackslist.append(
            tf.keras.callbacks.TensorBoard(log_dir=log_dir, histogram_freq=0, batch_size=32,
                                           write_graph=True, write_grads=False, write_images=False,
                                           update_freq=update_freq)
        )
    t0 = time.time()
    history = NN_tmp.fit(X_train, y_train, epochs=ep, batch_size=bs,
                         validation_data=(X_test, y_test),
                         callbacks=callbackslist,
                         verbose=train_verbose)
    print("training runtime: ", time.time() - t0)
    if restore_best:
        NN_tmp.load_weights(save_path)

    if plot_history:
        print(ns, lr, np.min(history.history['loss']), np.min(history.history['val_loss']))
        plt.plot(history.history['loss'])
        plt.plot(history.history['val_loss'])
        plt.title('MSE loss over epoches')
        plt.legend(['train', 'val'])
        plt.yscale('log')
        plt.show()

    if show_Elab:
        y_train_pred = NN_tmp.predict(X_train)
        y_test_pred = NN_tmp.predict(X_test)
        plot_E_lab(y_train, y_train_pred)
        plot_E_lab(y_test, y_test_pred)

    if show_ED:
        y_train_pred = NN_tmp.predict(X_train)
        y_test_pred = NN_tmp.predict(X_test)
        show_ED_fn(y_train, y_train_pred)
        show_ED_fn(y_test, y_test_pred)

    return NN_tmp, history


def floatInList(a_float, a_list):
    for tmp in a_list:
        if np.abs(a_float - tmp) < 1e-6:
            return True
    return False


class MyHistory:
    history = {}
    def __init__(self, model_filepath):
        history_train_loss_fname = model_filepath[:-3] + "_trainLoss.txt"
        history_valid_loss_fname = model_filepath[:-3] + "_validLoss.txt"
        history_lr_fname = model_filepath[:-3] + "_lr.txt"
        if os.path.exists(history_train_loss_fname):
            self.history['loss'] = np.loadtxt(history_train_loss_fname)
            self.history['val_loss'] = np.loadtxt(history_valid_loss_fname)
            self.history['lr'] = np.loadtxt(history_lr_fname)
        else:
            self.history['loss'] = np.array([0.])
            self.history['val_loss'] = np.array([0.])
            self.history['lr'] = np.array([0.])
            print("!" * 50)
            print("  >>>> [CAUTION] loss log file not found! Fake as only one epoch with loss=0. This makes xVal summary not accurate.")
            print("!" * 50)


class MyModelCheckpoint(keras.callbacks.Callback):
    def __init__(self, filepath, save_since=0, save_interval=500):
        super(MyModelCheckpoint, self).__init__()
        self.filepath = filepath
        self.save_since = save_since
        self.save_interval = save_interval

    def on_epoch_end(self, epoch, logs=None):
        if epoch == self.save_since or \
                (epoch > self.save_since and (epoch - self.save_since) % self.save_interval == 0):
            weights_fname = self.filepath[:-3] + "__epoch{}".format(epoch) + self.filepath[-3:]
            self.model.save_weights(weights_fname)
            print("epoch", epoch, ". Saved model weights to disk at", weights_fname)


class MyEarlyStopping(keras.callbacks.Callback):
    """Stop training when the loss is at its min, i.e. the loss stops decreasing.

  Arguments:
      patience: Number of epochs to wait after min has been hit. After this
      number of no improvement, training stops.
  """

    def __init__(self, patience=2000, train_loss_weight=0., filepath=None, dbg=False,
                 xx=None, yy=None):
        super(MyEarlyStopping, self).__init__()
        self.patience = patience
        self.train_loss_weight = train_loss_weight
        # filepath will be passed into get_best_lr_xVal_v2() for model saving & split/loss logging,
        # and inside get_best_lr_xVal_v2() it will be appended with do & lr info.
        self.filepath = filepath

        self.dbg = dbg
        self.xx = xx
        self.yy = yy

        # init: best_weights to store the weights at which the minimum loss occurs.
        self.best_weights = None
        self.best_weights_validLossOnly = None

    def on_train_begin(self, logs=None):
        #print("on_train_begin at", datetime.now())
        # The number of epoch it has waited when loss is no longer minimum.
        self.wait = 0
        # The epoch the training stops at.
        self.stopped_epoch = 0
        # Initialize the best as infinity.
        self.best = np.Inf
        self.best_epoch = 0
        self.best_validLossOnly = np.Inf
        self.best_epoch_validLossOnly = 0
        # printe weights for dbg
        dbg_ = False
        if dbg_:
            #print("on_train_begin")
            for layer in self.model.layers:
                if layer.name.startswith('dense'):
                    layer_weights = layer.get_weights()[0]
                    out_tmp = np.dot(np.ones([1, layer_weights.shape[0]]), layer_weights)
                    #print(layer.name, layer.units, layer_weights.shape, layer_weights.mean(), out_tmp.shape, out_tmp.mean())


    def on_epoch_end(self, epoch, logs=None):
        '''
        layer_dict = dict(zip([layer.name for layer in self.model.layers], self.model.layers))
        print(epoch, "(epoch end) dense_2 weights mean:", layer_dict['dense_2'].trainable_weights[0].numpy().mean(axis=0))
        print(epoch, "(epoch begin) dense_1 weights mean:", layer_dict['dense_1'].trainable_weights[0].numpy().mean(axis=0))
        print(epoch, "(epoch end) dense weights mean:", layer_dict['dense'].trainable_weights[0].numpy().mean(axis=0))
        '''
        #print(epoch, "on_epoch_end at", datetime.now())
        if self.dbg:
            pass
        current_train_loss = logs.get("loss")
        current_valid_loss = logs.get("val_loss")
        current = self.train_loss_weight * current_train_loss + current_valid_loss
        if epoch == 0 or np.less(current_valid_loss, self.best_validLossOnly):
            self.best_validLossOnly = current_valid_loss
            self.best_weights_validLossOnly = self.model.get_weights()
            self.best_epoch_validLossOnly = epoch
        if epoch == 0 or np.less(current, self.best):
            self.best = current
            self.wait = 0
            # Record the best weights if current results is better (less).
            self.best_weights = self.model.get_weights()
            self.best_epoch = epoch
        else:
            self.wait += 1
            if self.wait >= self.patience:
                self.stopped_epoch = epoch
                self.model.stop_training = True

    def on_train_end(self, logs=None):
        #print("on_train_end at", datetime.now())
        #print("[DEBUG] Restoring model weights from the end of the best epoch (%05d)." % (self.best_epoch_validLossOnly + 1))
        #self.model.set_weights(self.best_weights_validLossOnly)
        #self.model.save_weights(self.filepath[:-3] + '__validLossOnly' + self.filepath[-3:])
        #print("[DEBUG] Saved model weights to disk at", self.filepath[:-3] + '__validLossOnly' + self.filepath[-3:])

        #print("Restoring model weights from the end of the best epoch (%05d)." % (self.best_epoch + 1))
        self.model.set_weights(self.best_weights)
        # save model weights
        self.model.save_weights(self.filepath)
        #print("Saved model weights to disk at", self.filepath)

        if self.stopped_epoch > 0:
            pass
            #print("Epoch %05d: early stopping (best epoch is %05d)" % (self.stopped_epoch + 1, self.best_epoch + 1))
        #print("on_train_end finishes at", datetime.now())


def my_train(X,
                        y,
                        y_ref = None,
                        y_ref_plus_offset = None,
                        CMYKG2cmykg=False,
                        cmykg_noMul=False,
                        cmykg_addOne=False,
                        cmykg_sigmoidAfterMul=False,
                        cmykg_clipAfterMul=False,
                        weight_all = None,
                        weight_train = None,
                        last_round_weight_train = None,
                        last_round_error_arr = None,
                        weight_thre = 1e-3,
                        ns=[1500, 1500, 1500], n_out=None, activation='relu', activation_final='linear', lrelu_alpha=0.01,
                        sigmoid_hscale_factor=1.0, sigmoid_hshift_amount=0.0, trainNum_for_sigAmount=None,
                        hlayer_split=-2, ns2=None, n_out2=None,
                        scales=None,
                        BN=True, BN_on_input=False,
                        kernel_initializer="glorot_normal",
                        kernel_initializer_final="glorot_normal",
                        kernel_init_seed=None,
                        clipnorm=None,
                        dropout=0.0, dropout_on_input=0.0,
                        reg_l1=0.0, reg_l2=0.0, loss=None,
                        init_lr=1e-3, bs=None, ep=100,
                        metricOn='train',
                        callbacks=[],
                        show_history=True,
                        show_metric=False,
                        show_metric_avg=True,
                        verbose=0,
                        max_start_valid_error=None,
                        search_lr=False,
                        lr_list_besides_orig=None,
                        xval_seed=None,
                        xval_ratio=0.8,
                        xval_ratioForValSet=0.1,
                        xval_ratioForTestSet=None,
                        xval_lastNforTest=None,
                        idx_to_skip=None,
                        idx_dedicated_for_training=None,
                        test_set=None,
                        keep_train_amount=True,
                        use_all_for_training=False,
                        stratified_sampling=False,
                        stratified_base=None,
                        use_trainingSet_as_validSet=False,
                        idx_hiPrio_for_P=None,
                        distance_based_P=True,
                        set_sameSeed_P=False,
                        set_seed_trainValidSplit=False,
                        idx_pool_for_valid=None,
                        gamma_stratified_sampling=False,
                        ensure_strata_balance=False,
                        deterministicSplit=False,
                        deterministicSplitMode=1,
                        xval_folds=1,
                        xval_polyD=7,
                        xval_saveAllModels=False,
                        xval_reuseExistingModels=False,
                        xval_reuseExistingModels_only1stfold=False,
                        xval_assertIndex=False,
                        xval_modelStructureToLoad=None,
                        xval_modelWeightsToLoad=None,
                        old_way_to_load_base_model=False,
                        base_net_direct_use = False,
                        xval_layerToTrainSince=None,
                        xval_layerToUse=None,
                        xval_addOneMoreLayer=False,
                        xval_polyFeat=True,
                        xval_evalOnly=False,
                        xval_folds_evalOnly=1,
                        xval_specialCase=None,
                        xval_rndWeight=None,
                        model_outputs='specAndAlpha',
                        lossOrElab=1,
                        folds_not_counted_into_summary=None,
                        folds_not_use_sHshiftAmnt_of_allData=None,
                        always_save_sigShiftAmnt=False,
                        lrs_must_included=None,
                        reuse_model_saved_on_min_valid_loss=False,
                        run_eagerly=False,
                        interp_points=None,
                        interp_values=None,
                        ensemble_thr_folds=False,
                        ensemble_pThreshold=0):

    #print(datetime.now())

    assert(X.shape[0] == y.shape[0])
    if last_round_error_arr is not None:
        assert(X.shape[0] == last_round_error_arr.shape[0])
        assert(len(last_round_error_arr.shape) == 1)
    if interp_values is not None:
        assert(len(interp_points) == len(interp_values.shape) - 1)
    ################  backup  starts #################
    bs_bak = bs
    sigmoid_hshift_amount_bak = sigmoid_hshift_amount
    sigmoid_hshift_amount_ofAllData = inv_sigmoid(y.mean(axis=0)) if activation_final == "sigmoid_hshift" else None
    weight_train_bak = weight_train
    ################  backup  ends #################

    lr_list = []
    model_filepath_list = []
    train_loss_list = []
    valid_loss_list = []
    train_avgElab_list = []
    valid_avgElab_list = []
    test_avgElab_list = []
    metricOn_list = train_loss_list if metricOn == 'train' else valid_loss_list
    loss_best = np.inf
    train_loss_atBest = None
    valid_loss_atBest = None
    test_loss_atBest = None
    NN_tmp_atBest = None

    model_structure_saved = False if not xval_reuseExistingModels else True
    terminate_searching = False
    lr = init_lr
    cnt = 0

    train_losses, valid_losses = [], []
    metric_dict_train_list = []
    metric_dict_valid_list = []
    metric_dict_test_list = []
    t0 = time.time()
    #print('lr=', lr)
    lr_list.append(lr)
    # use the same seed before cross-val so that all combinations of {do, lr} have the same data split, so that
    # the comparison between different combinations is fair because they are trained & evaluated with the same
    # data. The data split between folds within the same combination will be different, which is no problem.
    folds = xval_folds if not xval_evalOnly else xval_folds_evalOnly
    np.random.seed(xval_seed)
    seeds = np.random.choice(10000, folds)
    seeds[0] = xval_seed
    #print("seeds:", seeds)
    #print("Have set random seed to ", xval_seed)
    # init for ensemble learning
    y_test_pred_list = []
    data_content_for_ensemble = None
    idx_P, idx_testSet = None, None
    for i in range(xval_folds if not xval_evalOnly else xval_folds_evalOnly):
        sys.stdout.flush()
        t1 = time.time()
        ################  re-init  starts #################
        bs = bs_bak
        sigmoid_hshift_amount = sigmoid_hshift_amount_bak
        weight_train = weight_train_bak
        ################  re-init  ends #################
        print("\n")
        print(datetime.now())
        fold = i+1
        print("##### fold ", fold, "#####")
        sys.stdout.flush()
        __seed = seeds[0] if set_sameSeed_P else seeds[i]
        np.random.seed(__seed)
        #print("set seed to", __seed)
        seed_trainValidSplit = seeds[i] if set_seed_trainValidSplit else None

        # [CAUTION]: the callbacks passed into this func contains instances of each callback. These instance is mutable
        # through different folds, and the changes cummulate, for example the ModelCheckpoint callback tracks the best
        # val_loss through folds (i.e. 2nd fold takes val_loss of 1st fold as baseline to judge whether should save
        # model), so if 1st fold's val_loss is too good then the following folds will never updated model, and then
        # every fold loads the same model when evaluating!!!! To avoid this, must deepcopy the callbacks instead of
        # using them directly!
        callbacks_deepcopy = copy.deepcopy(callbacks)
        # delete old model
        model_filepath = None
        ModelCheckpoint_callback = [cb for cb in callbacks_deepcopy if isinstance(cb, keras.callbacks.ModelCheckpoint)]
        if len(ModelCheckpoint_callback) == 0:
            ModelCheckpoint_callback = [cb for cb in callbacks_deepcopy if isinstance(cb, MyEarlyStopping)]
        skipThisFold = False
        if len(ModelCheckpoint_callback) > 0:
            model_filepath = ModelCheckpoint_callback[0].filepath
            assert('/loss_func_asLoss/' not in model_filepath)
            if xval_saveAllModels:
                model_filepath += "_do{}_lr{:.6}_fold{}".format(dropout, lr, fold)
            model_filepath += ".h5"
            ModelCheckpoint_callback[0].filepath = model_filepath
            if os.path.exists(model_filepath):
                if xval_reuseExistingModels:
                    skipThisFold = True
                else:
                    os.remove(model_filepath)
                    #sys.exit("Model already exists:", model_filepath)

            MyModelCheckpoint_callback = [cb for cb in callbacks_deepcopy if isinstance(cb, MyModelCheckpoint)]
            if len(MyModelCheckpoint_callback) > 0:
                MyModelCheckpoint_callback[0].filepath = model_filepath

        ratio = xval_ratio
        ratio2 = xval_ratioForValSet
        ratio3 = xval_ratioForTestSet
        if model_outputs == 'lab':
            assert(y.shape[1] == 3)
        elif model_outputs == 'spec':
            assert(y.shape[1] == 31)
        elif model_outputs == 'specAndAlpha':
            assert(y.shape[1] == 32)

        cmykt_train_, cmykt_valid_, cmykt_test_, lab_train_, lab_valid_, lab_test_, idx_train, idx_valid, idx_test = data_split_v3(
            X, y, ratio, ratio2=ratio2, ratio3=ratio3, stratified_sampling=stratified_sampling, stratified_base=stratified_base, idx_to_skip=idx_to_skip, idx_dedicated_for_training=idx_dedicated_for_training, keep_train_amount=keep_train_amount, use_all_for_training=use_all_for_training, returnIdx=True, seed=None, seed_trainValidSplit=seed_trainValidSplit, idx_hiPrio_for_P=idx_hiPrio_for_P, distance_based_P=distance_based_P, idx_pool_for_valid=idx_pool_for_valid, gamma_stratified_sampling=gamma_stratified_sampling, ensure_strata_balance=ensure_strata_balance, deterministicSplit=deterministicSplit, deterministicSplitMode=deterministicSplitMode, test_set=test_set)
        if True: # and fold == 1:
            _P = np.concatenate([idx_train, idx_valid], axis=0)
            #print("P/train/valid/test idx lengths:", np.shape(_P)[0], np.shape(idx_train)[0], np.shape(idx_valid)[0], np.shape(idx_test)[0])
            #print("P/train/valid/test idx means:", np.mean(_P), np.mean(idx_train), np.mean(idx_valid), np.mean(idx_test))
            #print("train/valid/test idx last 3:", idx_train[-3:], idx_valid[-3:], idx_test[-3:])

        X_all, y_all = X, y
        X_train, y_train = cmykt_train_, lab_train_
        X_valid, y_valid = cmykt_valid_, lab_valid_
        X_test, y_test = cmykt_test_, lab_test_
        if fold == 1:
            pass
            #print("X shapes:", X_train.shape, X_valid.shape, X_test.shape, X.min(), X.max())
            #print("y shapes:", y_train.shape, y_valid.shape, y_test.shape, y.min(), y.max())
        if bs is None:
            bs = X_train.shape[0]
            #print("bs is change from None to", bs)
        elif bs < 1:
            assert(bs > 0)
            bs = X_train.shape[0] * bs  # take it as a fraction
            bs = np.ceil(bs).astype('int')  # ceiling to avoid "leave-one" mini-batch issue. Then cast to int
            #print("bs is change from {} to {}".format(bs_bak, bs))
        #print("batch size:", bs)

        n_in = X_train.shape[1]
        if n_out is None:
            n_out = y_train.shape[1]
        else:
            pass
            #print("Set to user-defined n_out:", n_out)

        if activation_final == "sigmoid_hshift":
            if sigmoid_hshift_amount is None:
                if trainNum_for_sigAmount is not None:
                    sigmoid_hshift_amount = inv_sigmoid(y_train[:trainNum_for_sigAmount].mean(axis=0))
                    #print(f"Use sigmoid_hshift_amount of the first {trainNum_for_sigAmount} training samples.")
                else:
                    sigmoid_hshift_amount = inv_sigmoid(y_train.mean(axis=0))
                    #print("Use sigmoid_hshift_amount of training data.")
        else:
            sigmoid_hshift_amount = 0.0

        K.clear_session()
        gc.collect()
        NN_tmp = build_model(n_in, ns, n_out,
                             CMYKG2cmykg=CMYKG2cmykg, cmykg_noMul=cmykg_noMul, cmykg_addOne=cmykg_addOne,
                             hlayer_split=hlayer_split, ns2=ns2, n_out2=n_out2,
                             activation=activation, activation_final=activation_final, scales=scales, BN=BN,
                             clipnorm=clipnorm,
                             kernel_initializer=kernel_initializer,
                             kernel_initializer_final=kernel_initializer_final,
                             kernel_init_seed=kernel_init_seed,
                             lrelu_alpha=lrelu_alpha,
                             sigmoid_hscale_factor=sigmoid_hscale_factor,
                             sigmoid_hshift_amount=sigmoid_hshift_amount,
                             dropout=dropout, reg_l1=reg_l1, reg_l2=reg_l2, loss=loss,
                             BN_on_input=BN_on_input, dropout_on_input=dropout_on_input,
                             run_eagerly=run_eagerly)
        K.set_value(NN_tmp.optimizer.lr, lr)
        if fold == 1:
            pass
            #print(NN_tmp.summary())
        if model_filepath is not None:
            model_structure_filepath = os.path.join(os.path.dirname(model_filepath), "model.json")
            if not model_structure_saved or not os.path.exists(model_structure_filepath):
                if not os.path.exists(os.path.dirname(model_structure_filepath)):
                    os.mkdir(os.path.dirname(model_structure_filepath))
                with open(model_structure_filepath, "w") as f:
                    f.write(NN_tmp.to_json())
                #print("Saved model structure as", model_structure_filepath)
                model_structure_saved = True
        if skipThisFold:
            history = MyHistory(model_filepath)  # to load saved history
        else:
            if weight_train is not None:
                get_statistics(weight_train, toPrint=True, returnDict=False, title="weight_train")
                for iii in range(10):
                    print(idx_train[iii], weight_train[iii])
            history = NN_tmp.fit(X_train, y_train, epochs=ep, batch_size=bs,
                             validation_data=(X_valid, y_valid),
                             sample_weight=weight_train,
                             callbacks=callbacks_deepcopy,
                             verbose=verbose)
        # load best intermediate model
        if model_filepath is not None:
            if reuse_model_saved_on_min_valid_loss == True:
                NN_tmp.load_weights(model_filepath[:-3] + "__validLossOnly.h5")
                #print("Loaded best intermediate model ", model_filepath[:-3] + "__validLossOnly.h5")
            else:
                NN_tmp.load_weights(model_filepath)
                #print("Loaded best intermediate model ", model_filepath)
        else:
            pass
            print("Use model from final epoch (not necessary the best model)")

        if show_metric or show_metric_avg:
            #y_all_pred = NN_tmp.predict(X_all)
            y_train_pred = NN_tmp.predict(X_train)
            y_valid_pred = NN_tmp.predict(X_valid)
            y_test_pred = NN_tmp.predict(X_test)
            #y_all_gt = y
            y_train_gt = y_train
            y_valid_gt = y_valid
            y_test_gt = y_test
            assert(model_outputs is not None)
            if not ensemble_thr_folds:
                metric_dict_train = get_metric_dict(y_train_pred, y_train_gt, data_content=model_outputs, toPrint=show_metric, title='train', cmykg_noMul=cmykg_noMul, cmykg_sigmoidAfterMul=cmykg_sigmoidAfterMul, cmykg_clipAfterMul=cmykg_clipAfterMul, interp_points=interp_points, interp_values=interp_values)
                metric_dict_valid = get_metric_dict(y_valid_pred, y_valid_gt, data_content=model_outputs, toPrint=show_metric, title='valid', cmykg_noMul=cmykg_noMul, cmykg_sigmoidAfterMul=cmykg_sigmoidAfterMul, cmykg_clipAfterMul=cmykg_clipAfterMul, interp_points=interp_points, interp_values=interp_values)
                metric_dict_test = get_metric_dict(y_test_pred, y_test_gt, data_content=model_outputs, toPrint=show_metric, title='test', cmykg_noMul=cmykg_noMul, cmykg_sigmoidAfterMul=cmykg_sigmoidAfterMul, cmykg_clipAfterMul=cmykg_clipAfterMul, interp_points=interp_points, interp_values=interp_values)
                metric_dict_train_list.append(metric_dict_train)
                metric_dict_valid_list.append(metric_dict_valid)
                metric_dict_test_list.append(metric_dict_test)
            else:
                _ = get_metric_dict(y_test_pred, y_test_gt, data_content=model_outputs, toPrint=show_metric, title='test', cmykg_noMul=cmykg_noMul, cmykg_sigmoidAfterMul=cmykg_sigmoidAfterMul, cmykg_clipAfterMul=cmykg_clipAfterMul, interp_points=interp_points, interp_values=interp_values)
                if model_outputs.startswith("cmykg"):
                    __y_test_pred, data_content_for_ensemble = cmykg_to_final_pred(y_test_pred, interp_points, interp_values, data_content=model_outputs, cmykg_sigmoidAfterMul=cmykg_sigmoidAfterMul, cmykg_clipAfterMul=cmykg_clipAfterMul)
                else:
                    __y_test_pred, data_content_for_ensemble = y_test_pred, model_outputs
                y_test_pred_list.append(__y_test_pred)
                # sanity check 1: P must be the same (though train/valid split can be different) across different folds, otherwise we actually use more data than size(P)
                if idx_P is None:
                    idx_P = set(np.concatenate([idx_train, idx_valid]))
                else:
                    assert(idx_P == set(np.concatenate([idx_train, idx_valid])))
                # sanity check 2: test set must be the same across different folds, otherwise the prediction averaging makes no sense
                if idx_testSet is None:
                    idx_testSet = set(idx_test)
                else:
                    assert(idx_testSet == set(idx_test))
        print("Runtime for this fold: {:.2f} sec".format(time.time() - t1))


    if show_metric_avg:
        if not ensemble_thr_folds:
            metric_dict_train_combined = summarize_metric_dict_list(metric_dict_train_list, toPrint=True, title="train", returnSummarizedDict=True)
            metric_dict_valid_combined = summarize_metric_dict_list(metric_dict_valid_list, toPrint=True, title="valid", returnSummarizedDict=True)
            metric_dict_test_combined = summarize_metric_dict_list(metric_dict_test_list, toPrint=True, title="test", returnSummarizedDict=True)
        else:
            print("\n")
            print(datetime.now())
            print('#' * 10 + " model averaging through folds " + '#' * 10)
            preds_test_arr = np.array(y_test_pred_list)
            #print("preds_test_arr shape:", preds_test_arr.shape)
            preds_test_combined = np.mean(preds_test_arr, axis=0)
            #print("preds_test_combined shape:", preds_test_combined.shape)
            _ = get_metric_dict(preds_test_combined, y_test_gt, data_content=data_content_for_ensemble, toPrint=True, title='test (ensemble)', cmykg_noMul=cmykg_noMul, cmykg_sigmoidAfterMul=cmykg_sigmoidAfterMul, cmykg_clipAfterMul=cmykg_clipAfterMul, interp_points=interp_points, interp_values=interp_values)
    print("\n")
    print(datetime.now())
    print("All {} folds are done. Total training runtime: {:.2f} sec".format(xval_folds, time.time() - t0))

